# Copyright (c) 2019 NTT.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils.excutils import save_and_reraise_exception
from oslo_utils import netutils
from oslo_utils import strutils

from blazar import context
from blazar.db import api as db_api
from blazar.db import exceptions as db_ex
from blazar.db import utils as db_utils
from blazar import exceptions
from blazar.manager import exceptions as manager_ex
from blazar.plugins import base
from blazar.plugins import floatingips as plugin
from blazar import status
from blazar.utils.openstack import neutron
from blazar.utils import plugins as plugins_utils


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

QUERY_TYPE_ALLOCATION = 'allocation'


class FloatingIpPlugin(base.BasePlugin):
    """Plugin for floating IP resource."""

    resource_type = plugin.RESOURCE_TYPE
    title = 'Floating IP Plugin'
    description = 'This plugin creates and assigns floating IPs.'
    query_options = {
        QUERY_TYPE_ALLOCATION: ['lease_id', 'reservation_id']
    }

    def check_params(self, values):
        if 'network_id' not in values:
            raise manager_ex.MissingParameter(param='network_id')

        if 'amount' not in values:
            raise manager_ex.MissingParameter(param='amount')

        if not strutils.is_int_like(values['amount']):
            raise manager_ex.MalformedParameter(param='amount')

        # required_floatingips param is an optional parameter
        fips = values.get('required_floatingips', [])
        if not isinstance(fips, list):
            manager_ex.MalformedParameter(param='required_floatingips')

        for ip in fips:
            if not (netutils.is_valid_ipv4(ip) or netutils.is_valid_ipv6(ip)):
                raise manager_ex.InvalidIPFormat(ip=ip)

    def _update_allocations(self, dates_before, dates_after, reservation_id,
                            reservation_status, fip_reservation, values):
        amount = int(values.get('amount', fip_reservation['amount']))
        fip_allocations = db_api.fip_allocation_get_all_by_values(
            reservation_id=reservation_id)
        allocs_to_remove = self._allocations_to_remove(
            dates_before, dates_after, fip_allocations, amount)

        if (allocs_to_remove and
                reservation_status == status.reservation.ACTIVE):
            raise manager_ex.CantUpdateFloatingIPReservation(
                msg="Cannot remove allocations from an active reservation")

        kept_fips = len(fip_allocations) - len(allocs_to_remove)
        fip_ids_to_add = []

        if kept_fips < amount:
            needed_fips = amount - kept_fips
            required_fips = values.get(
                'required_floatingips',
                fip_reservation['required_floatingips'])
            fip_ids_to_add = self._matching_fips(
                fip_reservation['network_id'], required_fips, needed_fips,
                dates_after['start_date'], dates_after['end_date'])

            if len(fip_ids_to_add) < needed_fips:
                raise manager_ex.NotEnoughFloatingIPAvailable()

        # Create new floating IPs if reservation is active
        created_fips = []
        if reservation_status == status.reservation.ACTIVE:
            ctx = context.current()
            fip_pool = neutron.FloatingIPPool(fip_reservation['network_id'])
            for fip_id in fip_ids_to_add:
                try:
                    fip = db_api.floatingip_get(fip_id)
                    LOG.debug('Creating floating IP %s for reservation %s',
                              fip['floating_ip_address'], reservation_id)
                    fip_pool.create_reserved_floatingip(
                        fip['subnet_id'], fip['floating_ip_address'],
                        ctx.project_id, reservation_id)
                    created_fips.append(fip['floating_ip_address'])
                except Exception as e:
                    for fip_address in created_fips:
                        fip_pool.delete_reserved_floatingip(fip_address)
                    err_msg = 'Failed to create floating IP: {}'.format(str(e))
                    raise manager_ex.NeutronClientError(err_msg)

        for fip_id in fip_ids_to_add:
            LOG.debug('Adding floating IP %s to reservation %s',
                      fip_id, reservation_id)
            db_api.fip_allocation_create({
                'floatingip_id': fip_id,
                'reservation_id': reservation_id})

        for allocation in allocs_to_remove:
            LOG.debug('Removing floating IP %s from reservation %s',
                      allocation['floatingip_id'], reservation_id)
            db_api.fip_allocation_destroy(allocation['id'])

    def _allocations_to_remove(self, dates_before, dates_after, allocs,
                               amount):
        """Find candidate floating IP allocations to remove."""
        allocs_to_remove = []

        for alloc in allocs:
            is_extension = (
                dates_before['start_date'] > dates_after['start_date'] or
                dates_before['end_date'] < dates_after['end_date'])

            if is_extension:
                reserved_periods = db_utils.get_reserved_periods(
                    alloc['floatingip_id'],
                    dates_after['start_date'],
                    dates_after['end_date'],
                    datetime.timedelta(seconds=1),
                    resource_type='floatingip')

                max_start = max(dates_before['start_date'],
                                dates_after['start_date'])
                min_end = min(dates_before['end_date'],
                              dates_after['end_date'])

                if not (len(reserved_periods) == 0 or
                        (len(reserved_periods) == 1 and
                         reserved_periods[0][0] == max_start and
                         reserved_periods[0][1] == min_end)):
                    allocs_to_remove.append(alloc)
                    continue

        allocs_to_keep = [a for a in allocs if a not in allocs_to_remove]

        if len(allocs_to_keep) > amount:
            allocs_to_remove.extend(
                allocs_to_keep[:(len(allocs_to_keep) - amount)])

        return allocs_to_remove

    def reserve_resource(self, reservation_id, values):
        """Create floating IP reservation."""
        self.check_params(values)

        required_fips = values.get('required_floatingips', [])
        amount = int(values['amount'])

        if len(required_fips) > amount:
            raise manager_ex.TooLongFloatingIPs()

        floatingip_ids = self._matching_fips(values['network_id'],
                                             required_fips,
                                             amount,
                                             values['start_date'],
                                             values['end_date'])

        floatingip_rsrv_values = {
            'reservation_id': reservation_id,
            'network_id': values['network_id'],
            'amount': amount
        }

        fip_reservation = db_api.fip_reservation_create(floatingip_rsrv_values)
        for fip_address in required_fips:
            fip_address_values = {
                'address': fip_address,
                'floatingip_reservation_id': fip_reservation['id']
            }
            db_api.required_fip_create(fip_address_values)
        for fip_id in floatingip_ids:
            db_api.fip_allocation_create({'floatingip_id': fip_id,
                                          'reservation_id': reservation_id})
        return fip_reservation['id']

    def update_reservation(self, reservation_id, values):
        """Update reservation."""
        reservation = db_api.reservation_get(reservation_id)
        lease = db_api.lease_get(reservation['lease_id'])
        dates_before = {'start_date': lease['start_date'],
                        'end_date': lease['end_date']}
        dates_after = {'start_date': values['start_date'],
                       'end_date': values['end_date']}
        fip_reservation = db_api.fip_reservation_get(
            reservation['resource_id'])

        if ('network_id' in values and
                values.get('network_id') != fip_reservation['network_id']):
            raise manager_ex.CantUpdateFloatingIPReservation(
                msg="Updating network_id is not supported")

        required_fips = fip_reservation['required_floatingips']
        if ('required_floatingips' in values and
                values['required_floatingips'] != required_fips and
                values['required_floatingips'] != []):
            raise manager_ex.CantUpdateFloatingIPReservation(
                msg="Updating required_floatingips is not supported except "
                    "with an empty list")

        self._update_allocations(dates_before, dates_after, reservation_id,
                                 reservation['status'], fip_reservation,
                                 values)
        updates = {}
        if 'amount' in values:
            updates['amount'] = values.get('amount')
        if updates:
            db_api.fip_reservation_update(fip_reservation['id'], updates)

        if ('required_floatingips' in values and
                values['required_floatingips'] != required_fips):
            db_api.required_fip_destroy_by_fip_reservation_id(
                fip_reservation['id'])
            for fip_address in values.get('required_floatingips'):
                fip_address_values = {
                    'address': fip_address,
                    'floatingip_reservation_id': fip_reservation['id']
                }
                db_api.required_fip_create(fip_address_values)

    def on_start(self, resource_id):
        fip_reservation = db_api.fip_reservation_get(resource_id)
        allocations = db_api.fip_allocation_get_all_by_values(
            reservation_id=fip_reservation['reservation_id'])

        ctx = context.current()
        fip_pool = neutron.FloatingIPPool(fip_reservation['network_id'])
        for alloc in allocations:
            fip = db_api.floatingip_get(alloc['floatingip_id'])
            fip_pool.create_reserved_floatingip(
                fip['subnet_id'], fip['floating_ip_address'],
                ctx.project_id, fip_reservation['reservation_id'])

    def on_end(self, resource_id):
        fip_reservation = db_api.fip_reservation_get(resource_id)
        allocations = db_api.fip_allocation_get_all_by_values(
            reservation_id=fip_reservation['reservation_id'])

        fip_pool = neutron.FloatingIPPool(fip_reservation['network_id'])
        for alloc in allocations:
            fip = db_api.floatingip_get(alloc['floatingip_id'])
            fip_pool.delete_reserved_floatingip(fip['floating_ip_address'])

    def allocation_candidates(self, values):
        self.check_params(values)

        required_fips = values.get('required_floatingips', [])
        amount = int(values['amount'])

        if len(required_fips) > amount:
            raise manager_ex.TooLongFloatingIPs()

        return self._matching_fips(values['network_id'], required_fips,
                                   amount, values['start_date'],
                                   values['end_date'])

    def _matching_fips(self, network_id, fip_addresses, amount,
                       start_date, end_date):
        filter_array = []
        start_date_with_margin = start_date - datetime.timedelta(
            minutes=CONF.cleaning_time)
        end_date_with_margin = end_date + datetime.timedelta(
            minutes=CONF.cleaning_time)

        fip_query = ["==", "$floating_network_id", network_id]
        filter_array = plugins_utils.convert_requirements(fip_query)

        fip_ids = []
        not_allocated_fip_ids = []
        allocated_fip_ids = []
        for fip in db_api.reservable_fip_get_all_by_queries(filter_array):
            if not db_api.fip_allocation_get_all_by_values(
                    floatingip_id=fip['id']):
                if fip['floating_ip_address'] in fip_addresses:
                    fip_ids.append(fip['id'])
                else:
                    not_allocated_fip_ids.append(fip['id'])
            elif db_utils.get_free_periods(
                    fip['id'],
                    start_date_with_margin,
                    end_date_with_margin,
                    end_date_with_margin - start_date_with_margin,
                    resource_type='floatingip'
            ) == [
                (start_date_with_margin, end_date_with_margin),
            ]:
                if fip['floating_ip_address'] in fip_addresses:
                    fip_ids.append(fip['id'])
                else:
                    allocated_fip_ids.append(fip['id'])

        if len(fip_ids) != len(fip_addresses):
            raise manager_ex.NotEnoughFloatingIPAvailable()

        fip_ids += not_allocated_fip_ids
        if len(fip_ids) >= amount:
            return fip_ids[:amount]

        fip_ids += allocated_fip_ids
        if len(fip_ids) >= amount:
            return fip_ids[:amount]

        raise manager_ex.NotEnoughFloatingIPAvailable()

    def validate_floatingip_params(self, values):
        marshall_attributes = set(['floating_network_id',
                                   'floating_ip_address'])
        missing_attr = marshall_attributes - set(values.keys())
        if missing_attr:
            raise manager_ex.MissingParameter(param=','.join(missing_attr))

    def create_floatingip(self, values):

        self.validate_floatingip_params(values)

        network_id = values.pop('floating_network_id')
        floatingip_address = values.pop('floating_ip_address')

        pool = neutron.FloatingIPPool(network_id)
        # validate the floating ip address is out of allocation_pools and
        # within its subnet cidr.
        try:
            subnet = pool.fetch_subnet(floatingip_address)
        except exceptions.BlazarException:
            with save_and_reraise_exception():
                LOG.info("Floating IP %s in network %s can't be used "
                         "for Blazar's resource.", floatingip_address,
                         network_id)

        floatingip_values = {
            'floating_network_id': network_id,
            'subnet_id': subnet['id'],
            'floating_ip_address': floatingip_address
        }

        floatingip = db_api.floatingip_create(floatingip_values)

        return floatingip

    def get(self, fip_id):
        return self.get_floatingip(fip_id)

    def get_floatingip(self, fip_id):
        fip = db_api.floatingip_get(fip_id)
        if fip is None:
            raise manager_ex.FloatingIPNotFound(floatingip=fip_id)
        return fip

    def list_floatingip(self):
        fips = db_api.floatingip_list()
        return fips

    def delete_floatingip(self, fip_id):
        fip = db_api.floatingip_get(fip_id)
        if fip is None:
            raise manager_ex.FloatingIPNotFound(floatingip=fip_id)

        allocations = db_api.fip_allocation_get_all_by_values(
            floatingip_id=fip_id)
        if allocations:
            msg = 'Floating IP id %s is allocated by reservations.' % fip_id
            LOG.info(msg)
            raise manager_ex.CantDeleteFloatingIP(floatingip=fip_id, msg=msg)
        try:
            db_api.floatingip_destroy(fip_id)
        except db_ex.BlazarDBException as e:
            raise manager_ex.CantDeleteFloatingIP(floatingip=fip_id,
                                                  msg=str(e))

    def list_allocations(self, query, detail=False):
        fip_id_list = [f['id'] for f in db_api.floatingip_list()]
        options = self.get_query_options(query, QUERY_TYPE_ALLOCATION)
        options['detail'] = detail
        fip_allocations = self.query_allocations(fip_id_list, **options)

        return [{"resource_id": fip, "reservations": allocs}
                for fip, allocs in fip_allocations.items()]

    def query_allocations(self, resource_id_list, detail=None, lease_id=None,
                          reservation_id=None):
        return self.query_fip_allocations(resource_id_list, detail=detail,
                                          lease_id=lease_id,
                                          reservation_id=reservation_id)

    def query_fip_allocations(self, fips, detail=None, lease_id=None,
                              reservation_id=None):
        """Return dict of host and its allocations.

        The list element forms
        {
          '-id': [
                       {
                         'lease_id': lease_id,
                         'id': reservation_id
                       },
                     ]
        }.
        """
        start = datetime.datetime.utcnow()
        end = datetime.date.max

        reservations = db_utils.get_reservation_allocations_by_fip_ids(
            fips, start, end, lease_id, reservation_id)
        fip_allocations = {fip: [] for fip in fips}

        for reservation in reservations:
            if not detail:
                del reservation['project_id']
                del reservation['lease_name']
                del reservation['status']

            for fip_id in reservation['floatingip_ids']:
                if fip_id in fip_allocations.keys():
                    fip_allocations[fip_id].append({
                        k: v for k, v in reservation.items()
                        if k != 'floatingip_ids'})

        return fip_allocations



# Copyright (c) 2017 NTT.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import datetime
import retrying

from novaclient import exceptions as nova_exceptions
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import strutils
from oslo_utils.strutils import bool_from_string

from blazar import context
from blazar.db import api as db_api
from blazar.db import utils as db_utils
from blazar.manager import exceptions as mgr_exceptions
from blazar.plugins import base
from blazar.plugins import instances as plugin
from blazar.plugins import oshosts
from blazar import status
from blazar.utils.openstack import exceptions as openstack_ex
from blazar.utils.openstack import nova
from blazar.utils.openstack import placement
from blazar.utils import plugins as plugins_utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

RESERVATION_PREFIX = 'reservation'
FLAVOR_EXTRA_SPEC = "aggregate_instance_extra_specs:" + RESERVATION_PREFIX
INSTANCE_DELETION_TIMEOUT = 10 * 60 * 1000  # 10 minutes

NONE_VALUES = ('None', 'none', None)
QUERY_TYPE_ALLOCATION = 'allocation'


class VirtualInstancePlugin(base.BasePlugin, nova.NovaClientWrapper):
    """Plugin for virtual instance resources."""

    resource_type = plugin.RESOURCE_TYPE
    title = 'Virtual Instance Plugin'
    query_options = {
        QUERY_TYPE_ALLOCATION: ['lease_id', 'reservation_id']
    }

    def __init__(self):
        super(VirtualInstancePlugin, self).__init__(
            username=CONF.os_admin_username,
            password=CONF.os_admin_password,
            user_domain_name=CONF.os_admin_user_domain_name,
            project_name=CONF.os_admin_project_name,
            project_domain_name=CONF.os_admin_project_domain_name)

        self.freepool_name = CONF.nova.aggregate_freepool_name
        self.monitor = oshosts.host_plugin.PhysicalHostMonitorPlugin()
        self.monitor.register_healing_handler(self.heal_reservations)
        self.placement_client = placement.BlazarPlacementClient()

    def filter_hosts_by_reservation(self, hosts, start_date, end_date,
                                    excludes):
        free = []
        non_free = []

        for host in hosts:
            reservations = db_utils.get_reservations_by_host_id(host['id'],
                                                                start_date,
                                                                end_date)

            if excludes:
                reservations = [r for r in reservations
                                if r['id'] not in excludes]

            if reservations == []:
                free.append({'host': host, 'reservations': []})
            elif not [r for r in reservations
                      if r['resource_type'] == oshosts.RESOURCE_TYPE]:
                non_free.append({'host': host, 'reservations': reservations})

        return free, non_free

    def max_usages(self, host, reservations):
        def resource_usage_by_event(event, resource_type):
            return event['reservation']['instance_reservation'][resource_type]

        events_list = []
        for r in reservations:
            fetched_events = db_api.event_get_all_sorted_by_filters(
                sort_key='time', sort_dir='asc',
                filters={'lease_id': r['lease_id']})
            events_list.extend([{'event': e, 'reservation': r}
                                for e in fetched_events])

        events_list.sort(key=lambda x: x['event']['time'])

        max_vcpus = max_memory = max_disk = 0
        current_vcpus = current_memory = current_disk = 0

        for event in events_list:
            if event['event']['event_type'] == 'start_lease':
                current_vcpus += resource_usage_by_event(event, 'vcpus')
                current_memory += resource_usage_by_event(event, 'memory_mb')
                current_disk += resource_usage_by_event(event, 'disk_gb')
                if max_vcpus < current_vcpus:
                    max_vcpus = current_vcpus
                if max_memory < current_memory:
                    max_memory = current_memory
                if max_disk < current_disk:
                    max_disk = current_disk
            elif event['event']['event_type'] == 'end_lease':
                current_vcpus -= resource_usage_by_event(event, 'vcpus')
                current_memory -= resource_usage_by_event(event, 'memory_mb')
                current_disk -= resource_usage_by_event(event, 'disk_gb')

        return max_vcpus, max_memory, max_disk

    def get_hosts_list(self, host_info, cpus, memory, disk):
        hosts_list = []
        host = host_info['host']
        reservations = host_info['reservations']
        max_cpus, max_memory, max_disk = self.max_usages(host,
                                                         reservations)
        used_cpus, used_memory, used_disk = (cpus, memory, disk)
        while (max_cpus + used_cpus <= host['vcpus'] and
               max_memory + used_memory <= host['memory_mb'] and
               max_disk + used_disk <= host['local_gb']):
            hosts_list.append(host)
            used_cpus += cpus
            used_memory += memory
            used_disk += disk
        return hosts_list

    def allocation_candidates(self, reservation):
        return self.pickup_hosts(None, reservation)['added']

    def list_allocations(self, query):
        hosts_id_list = [h['id'] for h in db_api.host_list()]
        options = self.get_query_options(query, QUERY_TYPE_ALLOCATION)

        hosts_allocations = self.query_allocations(hosts_id_list, **options)
        return [{"resource_id": host, "reservations": allocs}
                for host, allocs in hosts_allocations.items()]

    def query_allocations(self, hosts, lease_id=None, reservation_id=None):
        """Return dict of host and its allocations.

        The list element forms
        {
          'host-id': [
                       {
                         'lease_id': lease_id,
                         'id': reservation_id
                         'start_date': lease_start_date,
                         'end_date': lease_end_date,
                       },
                     ]
        }.
        """
        start = datetime.datetime.utcnow()
        end = datetime.date.max

        # To reduce overhead, this method only executes one query
        # to get the allocation information
        reservations = db_utils.get_reservation_allocations_by_host_ids(
            hosts, start, end, lease_id, reservation_id)
        host_allocs = {h: [] for h in hosts}
        attributes_to_copy = ["id", "lease_id", "start_date", "end_date"]
        for reservation in reservations:
            for host_id in reservation['host_ids']:
                if host_id in host_allocs.keys():
                    host_allocs[host_id].append({
                        k: v for k, v in reservation.items()
                        if k in attributes_to_copy})
        return host_allocs

    def query_available_hosts(self, cpus=None, memory=None, disk=None,
                              resource_properties=None,
                              start_date=None, end_date=None,
                              excludes_res=None):
        """Returns a list of available hosts for a reservation.

        The list is in the order of reserved hosts to free hosts.

        1. filter hosts that have a spec enough to accommodate the flavor
        2. categorize hosts into hosts with and without allocation
           at the reservation time frame
        3. filter out hosts used by physical host reservation from
           allocate_host
        4. filter out hosts that can't accommodate the flavor at the
           time frame because of other reservations
        """
        flavor_definitions = [
            'and',
            [">=", "$vcpus", str(cpus)],
            [">=", "$memory_mb", str(memory)],
            [">=", "$local_gb", str(disk)],
            ]

        filters = plugins_utils.convert_requirements(flavor_definitions)

        if resource_properties:
            filters += plugins_utils.convert_requirements(resource_properties)

        hosts = db_api.reservable_host_get_all_by_queries(filters)
        free_hosts, reserved_hosts = self.filter_hosts_by_reservation(
            hosts,
            start_date - datetime.timedelta(minutes=CONF.cleaning_time),
            end_date + datetime.timedelta(minutes=CONF.cleaning_time),
            excludes_res)

        available_hosts = []
        for host_info in (reserved_hosts + free_hosts):
            hosts_list = self.get_hosts_list(host_info, cpus, memory, disk)
            available_hosts.extend(hosts_list)

        return available_hosts

    def pickup_hosts(self, reservation_id, values):
        """Returns lists of host ids to add/remove.

        This function picks up available hosts, calculates the difference from
        old reservations and returns a dict of a list of host ids to add
        and remove keyed by "added" or "removed".

        Note that the lists allow duplicated host ids for `affinity=True`
        cases.

        :raises: NotEnoughHostsAvailable exception if there are not enough
                 hosts available for the request
        """
        req_amount = values['amount']
        affinity = bool_from_string(values['affinity'], default=None)

        query_params = {
            'cpus': values['vcpus'],
            'memory': values['memory_mb'],
            'disk': values['disk_gb'],
            'resource_properties': values['resource_properties'],
            'start_date': values['start_date'],
            'end_date': values['end_date']
            }

        old_allocs = db_api.host_allocation_get_all_by_values(
            reservation_id=reservation_id)
        if old_allocs:
            # This is a path for *update* reservation. Add the specific
            # query param not to consider resources reserved by existing
            # reservations to update
            query_params['excludes_res'] = [reservation_id]

        new_hosts = self.query_available_hosts(**query_params)

        old_host_id_list = [h['compute_host_id'] for h in old_allocs]
        candidate_id_list = [h['id'] for h in new_hosts]

        # Build `new_host_id_list`. Note that we'd like to pick up hosts in
        # the following order of priority:
        #  1. hosts reserved by the reservation to update
        #  2. hosts with reservations followed by hosts without reservations
        # Note that the `candidate_id_list` has already been ordered
        # satisfying the second requirement.
        if affinity:
            host_id_map = collections.Counter(candidate_id_list)
            available = {k for k, v in host_id_map.items() if v >= req_amount}
            if not available:
                raise mgr_exceptions.NotEnoughHostsAvailable()
            new_host_ids = set(old_host_id_list) & available
            if new_host_ids:
                # (priority 1) This is a path for update reservation. We pick
                # up a host from hosts reserved by the reservation to update.
                new_host_id = new_host_ids.pop()
            else:
                # (priority 2) This is a path both for update and for new
                # reservation. We pick up hosts with some other reservations
                # if possible and otherwise pick up hosts without any
                # reservation. We can do so by considering the order of the
                # `candidate_id_list`.
                for host_id in candidate_id_list:
                    if host_id in available:
                        new_host_id = host_id
                        break
            new_host_id_list = [new_host_id] * req_amount
        else:
            # Hosts that can accommodate but don't satisfy priority 1
            _, possible_host_list = plugins_utils.list_difference(
                old_host_id_list, candidate_id_list)
            # Hosts that satisfy priority 1
            new_host_id_list, _ = plugins_utils.list_difference(
                candidate_id_list, possible_host_list)
            if affinity is False:
                # Eliminate the duplication
                new_host_id_list = list(set(new_host_id_list))
            for host_id in possible_host_list:
                if (affinity is False) and (host_id in new_host_id_list):
                    # Eliminate the duplication
                    continue
                new_host_id_list.append(host_id)
            if len(new_host_id_list) < req_amount:
                raise mgr_exceptions.NotEnoughHostsAvailable()
            while len(new_host_id_list) > req_amount:
                new_host_id_list.pop()

        # Calculate the difference from the existing reserved host
        removed_host_ids, added_host_ids = plugins_utils.list_difference(
            old_host_id_list, new_host_id_list)

        return {'added': added_host_ids, 'removed': removed_host_ids}

    def _create_flavor(self, reservation_id, vcpus, memory, disk,
                       group_id=None):
        flavor_details = {
            'flavorid': reservation_id,
            'name': RESERVATION_PREFIX + ":" + reservation_id,
            'vcpus': vcpus,
            'ram': memory,
            'disk': disk,
            'is_public': False
            }
        reserved_flavor = self.nova.nova.flavors.create(**flavor_details)

        # Set extra specs to the flavor
        rsv_id_rc_format = reservation_id.upper().replace("-", "_")
        reservation_rc = "resources:CUSTOM_RESERVATION_" + rsv_id_rc_format
        extra_specs = {
            FLAVOR_EXTRA_SPEC: reservation_id,
            reservation_rc: "1"
            }
        if group_id is not None:
            extra_specs["affinity_id"] = group_id
        reserved_flavor.set_keys(extra_specs)

        return reserved_flavor

    def _create_resources(self, inst_reservation):
        reservation_id = inst_reservation['reservation_id']

        ctx = context.current()
        user_client = nova.NovaClientWrapper()

        flavor_args = {
            'reservation_id': reservation_id,
            'vcpus': inst_reservation['vcpus'],
            'memory': inst_reservation['memory_mb'],
            'disk': inst_reservation['disk_gb']
        }

        pool_metadata = {
            RESERVATION_PREFIX: reservation_id,
            'filter_tenant_id': ctx.project_id,
            }

        if inst_reservation['affinity'] is not None:
            reserved_group = user_client.nova.server_groups.create(
                RESERVATION_PREFIX + ':' + reservation_id,
                'affinity' if inst_reservation['affinity'] else 'anti-affinity'
                )
            flavor_args['group_id'] = reserved_group.id
            pool_metadata['affinity_id'] = reserved_group.id
        else:
            reserved_group = None

        reserved_flavor = self._create_flavor(**flavor_args)

        pool = nova.ReservationPool()
        agg = pool.create(name=reservation_id, metadata=pool_metadata)

        self.placement_client.create_reservation_class(reservation_id)

        return reserved_flavor, reserved_group, agg

    def cleanup_resources(self, instance_reservation):
        def check_and_delete_resource(client, id):
            try:
                client.delete(id)
            except nova_exceptions.NotFound:
                pass

        reservation_id = instance_reservation['reservation_id']

        server_group_id = instance_reservation['server_group_id']
        if server_group_id:
            check_and_delete_resource(self.nova.nova.server_groups,
                                      server_group_id)
        check_and_delete_resource(self.nova.nova.flavors, reservation_id)
        check_and_delete_resource(nova.ReservationPool(), reservation_id)

    def update_resources(self, reservation_id):
        """Updates reserved resources in Nova.

        This method updates reserved resources in Compute service. If the
        reservation is in active status, it adds new allocated hosts into
        a reserved aggregate. If the reservation is not started yet, it
        updates a reserved flavor.
        """
        reservation = db_api.reservation_get(reservation_id)

        if reservation['status'] == 'active':
            pool = nova.ReservationPool()

            # Dict of number of instances to reserve on a host keyed by the
            # host id
            allocation_map = collections.defaultdict(lambda: 0)
            for allocation in db_api.host_allocation_get_all_by_values(
                    reservation_id=reservation['id']):
                host_id = allocation['compute_host_id']
                allocation_map[host_id] += 1

            for host_id, num in allocation_map.items():
                host = db_api.host_get(host_id)
                try:
                    pool.add_computehost(
                        reservation['aggregate_id'],
                        host['service_name'], stay_in=True)
                except mgr_exceptions.AggregateAlreadyHasHost:
                    pass
                except nova_exceptions.ClientException:
                    err_msg = ('Fail to add host %s to aggregate %s.'
                               % (host, reservation['aggregate_id']))
                    raise mgr_exceptions.NovaClientError(err_msg)
                self.placement_client.update_reservation_inventory(
                    host['hypervisor_hostname'], reservation['id'], num)
        else:
            try:
                self.nova.nova.flavors.delete(reservation['id'])
                self._create_flavor(reservation['id'],
                                    reservation['vcpus'],
                                    reservation['memory_mb'],
                                    reservation['disk_gb'],
                                    reservation['server_group_id'])
            except nova_exceptions.ClientException:
                LOG.exception("Failed to update Nova resources "
                              "for reservation %s", reservation['id'])
                raise mgr_exceptions.NovaClientError()

    def _check_missing_reservation_params(self, values):
        marshall_attributes = set(['vcpus', 'memory_mb', 'disk_gb',
                                   'amount', 'affinity',
                                   'resource_properties'])
        missing_attr = marshall_attributes - set(values.keys())
        if missing_attr:
            raise mgr_exceptions.MissingParameter(param=','.join(missing_attr))

    def _validate_reservation_params(self, values):
        if 'amount' in values:
            try:
                values['amount'] = strutils.validate_integer(
                    values['amount'], "amount", 1, db_api.DB_MAX_INT)
            except ValueError as e:
                raise mgr_exceptions.MalformedParameter(str(e))

        if 'affinity' in values:
            if (values['affinity'] not in NONE_VALUES and
                    not strutils.is_valid_boolstr(values['affinity'])):
                raise mgr_exceptions.MalformedParameter(
                    param='affinity (must be a bool value or None)')

    def reserve_resource(self, reservation_id, values):
        self._check_missing_reservation_params(values)
        self._validate_reservation_params(values)

        hosts = self.pickup_hosts(reservation_id, values)

        instance_reservation_val = {
            'reservation_id': reservation_id,
            'vcpus': values['vcpus'],
            'memory_mb': values['memory_mb'],
            'disk_gb': values['disk_gb'],
            'amount': values['amount'],
            'affinity': bool_from_string(values['affinity'], default=None),
            'resource_properties': values['resource_properties']
            }
        instance_reservation = db_api.instance_reservation_create(
            instance_reservation_val)

        for host_id in hosts['added']:
            db_api.host_allocation_create({'compute_host_id': host_id,
                                          'reservation_id': reservation_id})

        try:
            flavor, group, pool = self._create_resources(instance_reservation)
        except nova_exceptions.ClientException:
            LOG.exception("Failed to create Nova resources "
                          "for reservation %s", reservation_id)
            self.cleanup_resources(instance_reservation)
            raise mgr_exceptions.NovaClientError()

        server_group_id = group.id if group is not None else None
        db_api.instance_reservation_update(instance_reservation['id'],
                                           {'flavor_id': flavor.id,
                                            'server_group_id': server_group_id,
                                            'aggregate_id': pool.id})

        return instance_reservation['id']

    def update_host_allocations(self, added, removed, reservation_id):
        allocations = db_api.host_allocation_get_all_by_values(
            reservation_id=reservation_id)

        removed_allocs = []
        for host_id in removed:
            for allocation in allocations:
                if allocation['compute_host_id'] == host_id:
                    removed_allocs.append(allocation['id'])
                    break

        # TODO(tetsuro): It would be nice to have something like
        # db_api.host_allocation_replace() to process the following
        # deletion and addition in *one* DB transaction.
        for alloc_id in removed_allocs:
            db_api.host_allocation_destroy(alloc_id)

        for added_host in added:
            db_api.host_allocation_create({'compute_host_id': added_host,
                                           'reservation_id': reservation_id})

    def update_reservation(self, reservation_id, new_values):
        """Updates an instance reservation with requested parameters.

        This method allows users to update an instance reservation under the
        following conditions.
        - If an instance reservation has not started yet
             - vcpus, memory_mb disk_gb and amount can be updateable unless
               Blazar can accommodate the new request.
        - If an instance reservation has already started
             - only amount is increasable.
        """
        self._validate_reservation_params(new_values)

        reservation = db_api.reservation_get(reservation_id)
        lease = db_api.lease_get(reservation['lease_id'])

        updatable = ['vcpus', 'memory_mb', 'disk_gb', 'affinity', 'amount',
                     'resource_properties']
        if (not any([k in updatable for k in new_values.keys()])
                and new_values['start_date'] >= lease['start_date']
                and new_values['end_date'] <= lease['end_date']):
            # no update because of just shortening the reservation time
            return

        if (reservation['status'] == 'active' and
                any([k in updatable[:-1] for k in new_values.keys()])):
            msg = "An active reservation only accepts to update amount."
            raise mgr_exceptions.InvalidStateUpdate(msg)

        if reservation['status'] == 'error':
            msg = "An error reservation doesn't accept an updating request."
            raise mgr_exceptions.InvalidStateUpdate(msg)

        if new_values.get('affinity', None):
            new_values['affinity'] = bool_from_string(new_values['affinity'],
                                                      default=None)

        for key in updatable:
            if key not in new_values:
                new_values[key] = reservation[key]

        changed_hosts = self.pickup_hosts(reservation_id, new_values)

        if (reservation['status'] == 'active'
                and len(changed_hosts['removed']) > 0):
            err_msg = ("Instance reservation doesn't allow to reduce/replace "
                       "reserved instance slots when the reservation is in "
                       "active status.")
            raise mgr_exceptions.CantUpdateParameter(err_msg)

        db_api.instance_reservation_update(
            reservation['resource_id'],
            {key: new_values[key] for key in updatable})

        self.update_host_allocations(changed_hosts['added'],
                                     changed_hosts['removed'],
                                     reservation_id)
        self.update_resources(reservation_id)

    def on_start(self, resource_id):
        ctx = context.current()
        instance_reservation = db_api.instance_reservation_get(resource_id)
        reservation_id = instance_reservation['reservation_id']

        # TODO(johngarbutt): create flavor after updating placement?
        # else we will race with automation looking for the flavor here
        try:
            self.nova.flavor_access.add_tenant_access(reservation_id,
                                                      ctx.project_id)
        except nova_exceptions.ClientException:
            LOG.info('Failed to associate flavor %(reservation_id)s '
                     'to project %(project_id)s',
                     {'reservation_id': reservation_id,
                      'project_id': ctx.project_id})
            raise mgr_exceptions.EventError()

        pool = nova.ReservationPool()

        # Dict of number of instances to reserve on a host keyed by the
        # host id
        allocation_map = collections.defaultdict(lambda: 0)
        for allocation in db_api.host_allocation_get_all_by_values(
                reservation_id=reservation_id):
            host_id = allocation['compute_host_id']
            allocation_map[host_id] += 1

        for host_id, num in allocation_map.items():
            host = db_api.host_get(host_id)
            pool.add_computehost(instance_reservation['aggregate_id'],
                                 host['service_name'], stay_in=True)
            self.placement_client.update_reservation_inventory(
                host['hypervisor_hostname'], reservation_id, num)

    def on_end(self, resource_id):
        instance_reservation = db_api.instance_reservation_get(resource_id)
        reservation_id = instance_reservation['reservation_id']
        ctx = context.current()

        try:
            self.nova.flavor_access.remove_tenant_access(
                reservation_id, ctx.project_id)
        except nova_exceptions.NotFound:
            pass

        hostnames = []
        allocations = db_api.host_allocation_get_all_by_values(
            reservation_id=reservation_id)
        for allocation in allocations:
            host = db_api.host_get(allocation['compute_host_id'])
            db_api.host_allocation_destroy(allocation['id'])
            hostnames.append(host['hypervisor_hostname'])

        for server in self.nova.servers.list(search_opts={
                'flavor': reservation_id,
                'all_tenants': 1}, detailed=False):
            try:
                self.nova.servers.delete(server=server)
            except nova_exceptions.NotFound:
                LOG.info("Could not find server '%s', may have been deleted "
                         "concurrently.", server.id)
            except Exception as e:
                LOG.exception("Failed to delete server '%s': %s.", server.id,
                              str(e))

        # We need to check the deletion is complete before deleting the
        # reservation inventory. See the bug #1813252 for details.
        if not self._check_server_deletion(reservation_id):
            LOG.error('Timed out while deleting servers on reservation %s',
                      reservation_id)
            raise mgr_exceptions.ServerDeletionTimeout()

        self.cleanup_resources(instance_reservation)

        for host_name in hostnames:
            try:
                self.placement_client.delete_reservation_inventory(
                    host_name, reservation_id)
            except openstack_ex.ResourceProviderNotFound:
                pass
        self.placement_client.delete_reservation_class(reservation_id)

    @retrying.retry(stop_max_delay=INSTANCE_DELETION_TIMEOUT,
                    wait_fixed=5000,  # 5 seconds interval
                    retry_on_result=lambda x: x is False)
    def _check_server_deletion(self, reservation_id):
        servers = self.nova.servers.list(search_opts={
            'flavor': reservation_id, 'all_tenants': 1}, detailed=False)
        if servers:
            LOG.info('Waiting to delete servers: %s ', servers)
            return False
        return True

    def heal_reservations(self, failed_resources, interval_begin,
                          interval_end):
        """Heal reservations which suffer from resource failures.

        :param failed_resources: failed resources
        :param interval_begin: start date of the period to heal.
        :param interval_end: end date of the period to heal.
        :return: a dictionary of {reservation id: flags to update}
                 e.g. {'de27786d-bd96-46bb-8363-19c13b2c6657':
                       {'missing_resources': True}}
        """
        reservation_flags = collections.defaultdict(dict)

        host_ids = [h['id'] for h in failed_resources]
        reservations = db_utils.get_reservations_by_host_ids(
            host_ids, interval_begin, interval_end)

        for reservation in reservations:
            if reservation['resource_type'] != plugin.RESOURCE_TYPE:
                continue

            if self._heal_reservation(reservation, host_ids):
                if reservation['status'] == status.reservation.ACTIVE:
                    reservation_flags[reservation['id']].update(
                        {'resources_changed': True})
            else:
                reservation_flags[reservation['id']].update(
                    {'missing_resources': True})

        return reservation_flags

    def _heal_reservation(self, reservation, host_ids):
        """Allocate alternative host(s) for the given reservation.

        :param reservation: A reservation that has allocations to change
        :param host_ids: Failed host ids
        :return: True if all the allocations in the given reservation
                 are successfully allocated
        """
        lease = db_api.lease_get(reservation['lease_id'])

        ret = True
        allocations = [
            alloc for alloc in reservation['computehost_allocations']
            if alloc['compute_host_id'] in host_ids]

        if reservation['affinity']:
            old_host_id = allocations[0]['compute_host_id']
            new_host_id = self._select_host(reservation, lease)

            self._pre_reallocate(reservation, old_host_id)

            if new_host_id is None:
                for allocation in allocations:
                    db_api.host_allocation_destroy(allocation['id'])
                LOG.warning('Could not find alternative host for '
                            'reservation %s (lease: %s).',
                            reservation['id'], lease['name'])
                ret = False
            else:
                for allocation in allocations:
                    db_api.host_allocation_update(
                        allocation['id'], {'compute_host_id': new_host_id})
                self._post_reallocate(
                    reservation, lease, new_host_id, len(allocations))

        else:
            new_host_ids = []
            for allocation in allocations:
                old_host_id = allocation['compute_host_id']
                new_host_id = self._select_host(reservation, lease)

                self._pre_reallocate(reservation, old_host_id)

                if new_host_id is None:
                    db_api.host_allocation_destroy(allocation['id'])
                    LOG.warning('Could not find alternative host for '
                                'reservation %s (lease: %s).',
                                reservation['id'], lease['name'])
                    ret = False
                    continue

                db_api.host_allocation_update(
                    allocation['id'], {'compute_host_id': new_host_id})
                new_host_ids.append(new_host_id)

            for new_host, num in collections.Counter(new_host_ids).items():
                self._post_reallocate(reservation, lease, new_host, num)

        return ret

    def _select_host(self, reservation, lease):
        """Returns the alternative host id or None if not found."""
        values = {}
        values['start_date'] = max(datetime.datetime.utcnow(),
                                   lease['start_date'])
        values['end_date'] = lease['end_date']
        specs = ['vcpus', 'memory_mb', 'disk_gb', 'affinity', 'amount',
                 'resource_properties']
        for key in specs:
            values[key] = reservation[key]
        try:
            changed_hosts = self.pickup_hosts(reservation['id'], values)
        except mgr_exceptions.NotEnoughHostsAvailable:
            return None
        # We should get at least one host to add because the old host can't
        # be in the candidates.
        return changed_hosts['added'][0]

    def _pre_reallocate(self, reservation, host_id):
        """Delete the reservation inventory/aggregates for the host."""
        pool = nova.ReservationPool()
        # Remove the failed host from the aggregate.
        if reservation['status'] == status.reservation.ACTIVE:
            host = db_api.host_get(host_id)
            pool.remove_computehost(reservation['aggregate_id'],
                                    host['service_name'])
            try:
                self.placement_client.delete_reservation_inventory(
                    host['hypervisor_hostname'], reservation['id'])
            except openstack_ex.ResourceProviderNotFound:
                pass

    def _post_reallocate(self, reservation, lease, host_id, num):
        """Add the reservation inventory/aggregates for the host."""
        pool = nova.ReservationPool()
        if reservation['status'] == status.reservation.ACTIVE:
            # Add the alternative host into the aggregate.
            new_host = db_api.host_get(host_id)
            pool.add_computehost(reservation['aggregate_id'],
                                 new_host['service_name'],
                                 stay_in=True)
            # Here we use "additional=True" not to break the existing
            # inventory(allocations) on the new host
            self.placement_client.update_reservation_inventory(
                new_host['hypervisor_hostname'], reservation['id'], num,
                additional=True)
        LOG.warning('Resource changed for reservation %s (lease: %s).',
                    reservation['id'], lease['name'])

    def _get_extra_capabilities(self, host_id):
        extra_capabilities = {}
        raw_extra_capabilities = (
            db_api.host_extra_capability_get_all_per_host(host_id))
        for capability, capability_name in raw_extra_capabilities:
            key = capability_name
            extra_capabilities[key] = capability.capability_value
        return extra_capabilities

    def get(self, host_id):
        host = db_api.host_get(host_id)
        extra_capabilities = self._get_extra_capabilities(host_id)
        if host is not None and extra_capabilities:
            res = host.copy()
            res.update(extra_capabilities)
            return res
        else:
            return host



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import uuid as uuidgen

from keystoneauth1 import session
from keystoneauth1 import token_endpoint
from novaclient import client as nova_client
from novaclient import exceptions as nova_exception
from novaclient.v2 import servers
from oslo_config import cfg
from oslo_log import log as logging

from blazar import context
from blazar.manager import exceptions as manager_exceptions
from blazar.plugins import oshosts
from blazar.utils.openstack import base


nova_opts = [
    cfg.StrOpt('endpoint_type',
               default='internal',
               choices=['public', 'admin', 'internal'],
               help='Type of the nova endpoint to use. This endpoint will be '
                    'looked up in the keystone catalog and should be one of '
                    'public, internal or admin.'),
    cfg.StrOpt('nova_client_version',
               default='2',
               deprecated_group='DEFAULT',
               help='Novaclient version'),
    cfg.StrOpt('compute_service',
               default='compute',
               deprecated_group='DEFAULT',
               help='Nova name in keystone'),
    cfg.StrOpt('image_prefix',
               default='reserved_',
               deprecated_group='DEFAULT',
               help='Prefix for VM images if you want to create snapshots'),
    cfg.StrOpt('aggregate_freepool_name',
               default='freepool',
               deprecated_group=oshosts.RESOURCE_TYPE,
               help='Name of the special aggregate where all hosts '
                    'are candidate for physical host reservation'),
    cfg.StrOpt('project_id_key',
               default='blazar:project',
               deprecated_group=oshosts.RESOURCE_TYPE,
               help='Aggregate metadata value for key matching project_id'),
    cfg.StrOpt('blazar_owner',
               default='blazar:owner',
               deprecated_group=oshosts.RESOURCE_TYPE,
               help='Aggregate metadata key for knowing owner project_id'),
    cfg.BoolOpt('az_aware',
                default=True,
                help='A flag to store original availability zone')
]


CONF = cfg.CONF
CONF.register_opts(nova_opts, group='nova')
CONF.import_opt('identity_service', 'blazar.utils.openstack.keystone')
LOG = logging.getLogger(__name__)


class BlazarNovaClient(object):
    def __init__(self, **kwargs):
        """Description

        BlazarNovaClient can be used in two ways: from context or kwargs.

        :param version: service client version which we will use
        :type version: str

        :param ctx: request context
        :type ctx: context object

        :param auth_token: keystone auth token
        :type auth_token: str

        :param endpoint_override: endpoint url which we will use
        :type endpoint_override: str

        :param username: username to use with nova client
        :type username: str

        :param password: password to use with nova client
        :type password: str

        :param user_domain_name: domain name of the user
        :type user_domain_name: str

        :param project_name: project name to use with nova client
        :type project_name: str

        :param project_domain_name: domain name of the project
        :type project_domain_name: str

        :param auth_url: keystone url to authenticate against
        :type auth_url: str
        """

        ctx = kwargs.pop('ctx', None)
        auth_token = kwargs.pop('auth_token', None)
        endpoint_override = kwargs.pop('endpoint_override', None)
        version = kwargs.pop('version', CONF.nova.nova_client_version)
        username = kwargs.pop('username', None)
        password = kwargs.pop('password', None)
        user_domain_name = kwargs.pop('user_domain_name', None)
        project_name = kwargs.pop('project_name', None)
        project_domain_name = kwargs.pop('project_domain_name', None)
        auth_url = kwargs.pop('auth_url', None)

        if ctx is None:
            try:
                ctx = context.current()
            except RuntimeError:
                pass
        if ctx is not None:
            auth_token = auth_token or ctx.auth_token
            endpoint_override = endpoint_override or \
                base.url_for(ctx.service_catalog,
                             CONF.nova.compute_service,
                             endpoint_interface=CONF.nova.endpoint_type,
                             os_region_name=CONF.os_region_name)
            auth_url = base.url_for(ctx.service_catalog, CONF.identity_service,
                                    CONF.endpoint_type,
                                    os_region_name=CONF.os_region_name)
            kwargs.setdefault('global_request_id', ctx.global_request_id)

        if auth_url is None:
            auth_url = "%s://%s:%s" % (CONF.os_auth_protocol,
                                       base.get_os_auth_host(CONF),
                                       CONF.os_auth_port)
            if CONF.os_auth_prefix:
                auth_url += "/%s" % CONF.os_auth_prefix

        if username:
            kwargs.setdefault('username', username)
            kwargs.setdefault('password', password)
            kwargs.setdefault('project_name', project_name)
            kwargs.setdefault('auth_url', auth_url)

            if "v2.0" not in auth_url:
                kwargs.setdefault('project_domain_name', project_domain_name)
                kwargs.setdefault('user_domain_name', user_domain_name)

            if CONF.cafile:
                kwargs.setdefault('cacert', CONF.cafile)
        else:
            auth = token_endpoint.Token(endpoint_override,
                                        auth_token)
            sess_kwargs = dict(
                auth=auth
            )
            if CONF.cafile:
                sess_kwargs.update(verify=CONF.cafile)
            sess = session.Session(**sess_kwargs)
            kwargs.setdefault('session', sess)

        kwargs.setdefault('endpoint_type', CONF.nova.endpoint_type + 'URL')
        kwargs.setdefault('endpoint_override', endpoint_override)
        kwargs.setdefault('version', version)
        self.nova = nova_client.Client(**kwargs)

        self.nova.servers = ServerManager(self.nova)

        self.exceptions = nova_exception

    def __getattr__(self, name):
        return getattr(self.nova, name)


class ServerManager(servers.ServerManager):

    def create_image(self, server, image_name=None, metadata=None):
        """Snapshot a server."""
        if image_name is None:
            image_name = CONF.nova.image_prefix + server.name
        return super(ServerManager, self).create_image(server,
                                                       image_name=image_name,
                                                       metadata=metadata)


class NovaClientWrapper(object):
    def __init__(self, username=None, password=None, user_domain_name=None,
                 project_name=None, project_domain_name=None):
        self.username = username
        self.password = password
        self.user_domain_name = user_domain_name
        self.project_name = project_name
        self.project_domain_name = project_domain_name

    @property
    def nova(self):
        nova = BlazarNovaClient(username=self.username,
                                password=self.password,
                                user_domain_name=self.user_domain_name,
                                project_name=self.project_name,
                                project_domain_name=self.project_domain_name)
        return nova


class ReservationPool(NovaClientWrapper):
    def __init__(self):
        super(ReservationPool, self).__init__(
            username=CONF.os_admin_username,
            password=CONF.os_admin_password,
            user_domain_name=CONF.os_admin_user_domain_name,
            project_name=CONF.os_admin_project_name,
            project_domain_name=CONF.os_admin_project_domain_name)

        self.config = CONF.nova
        self.freepool_name = self.config.aggregate_freepool_name

    def get_aggregate_from_name_or_id(self, aggregate_obj):
        """Return an aggregate by name or an id."""

        aggregate = None
        agg_id = None
        try:
            agg_id = int(aggregate_obj)
        except (ValueError, TypeError):
            if hasattr(aggregate_obj, 'id') and aggregate_obj.id:
                # pool is an aggregate
                agg_id = aggregate_obj.id

        if agg_id is not None:
            try:
                aggregate = self.nova.aggregates.get(agg_id)
            except nova_exception.NotFound:
                aggregate = None
        else:
            # FIXME(scroiset): can't get an aggregate by name
            # so iter over all aggregate and check for the good one
            all_aggregates = self.nova.aggregates.list()
            for agg in all_aggregates:
                if aggregate_obj == agg.name:
                    aggregate = agg
        if aggregate:
            return aggregate
        else:
            raise manager_exceptions.AggregateNotFound(pool=aggregate_obj)

    @staticmethod
    def _generate_aggregate_name():
        return str(uuidgen.uuid4())

    def create(self, name=None, az=None, metadata=None):
        """Create a Pool (an Aggregate) with or without Availability Zone.

        By default expose to user the aggregate with an Availability Zone.
        Return an aggregate or raise a nova exception.

        """

        name = name or self._generate_aggregate_name()

        LOG.debug('Creating pool aggregate: %(name)s with Availability Zone '
                  '%(az)s', {'name': name, 'az': az})
        agg = self.nova.aggregates.create(name, az)

        try:
            ctx = context.current()
            project_id = ctx.project_id
        except RuntimeError:
            e = manager_exceptions.ProjectIdNotFound()
            LOG.error(str(e))
            raise e

        if metadata:
            metadata[self.config.blazar_owner] = project_id
        else:
            metadata = {self.config.blazar_owner: project_id}
        self.nova.aggregates.set_metadata(agg, metadata)

        return agg

    def delete(self, pool, force=True):
        """Delete an aggregate.

        pool can be an aggregate name or an aggregate id.
        Remove all hosts before delete aggregate (default).
        If force is False, raise exception if at least one
        host is attached to.

        """

        try:
            agg = self.get_aggregate_from_name_or_id(pool)
        except manager_exceptions.AggregateNotFound:
            LOG.warning("Aggregate '%s' not found, skipping deletion", pool)
            return

        hosts = agg.hosts
        if len(hosts) > 0 and not force:
            raise manager_exceptions.AggregateHaveHost(name=agg.name,
                                                       hosts=agg.hosts)
        try:
            freepool_agg = self.get(self.freepool_name)
        except manager_exceptions.AggregateNotFound:
            raise manager_exceptions.NoFreePool()
        for host in hosts:
            LOG.debug("Removing host '%(host)s' from aggregate '%(id)s')",
                      {'host': host, 'id': agg.id})
            self.nova.aggregates.remove_host(agg.id, host)

            if freepool_agg.id != agg.id and host not in freepool_agg.hosts:
                self.nova.aggregates.add_host(freepool_agg.id, host)

        self.nova.aggregates.delete(agg.id)

    def get_all(self):
        """Return all aggregate."""

        return self.nova.aggregates.list()

    def get(self, pool):
        """return details for aggregate pool or raise AggregateNotFound."""

        return self.get_aggregate_from_name_or_id(pool)

    def get_computehosts(self, pool):
        """Return a list of compute host names for an aggregate."""

        try:
            agg = self.get_aggregate_from_name_or_id(pool)
            return agg.hosts
        except manager_exceptions.AggregateNotFound:
            return []

    def add_computehost(self, pool, hosts, stay_in=False):
        """Add compute host(s) to an aggregate.

        Each host must exist and be in the freepool, otherwise raise an error.

        :param pool: Name or UUID of the pool to rattach the host
        :param hosts: Names (not UUID) of hosts to associate
        :type host: str or list of str

        Return the related aggregate.
        Raise an aggregate exception if something wrong.
        """

        if not isinstance(hosts, list):
            hosts = [hosts]

        added_hosts = []
        removed_hosts = []
        agg = self.get_aggregate_from_name_or_id(pool)

        try:
            freepool_agg = self.get(self.freepool_name)
        except manager_exceptions.AggregateNotFound:
            raise manager_exceptions.NoFreePool()

        try:
            for host in hosts:
                if freepool_agg.id != agg.id and not stay_in:
                    if host not in freepool_agg.hosts:
                        raise manager_exceptions.HostNotInFreePool(
                            host=host, freepool_name=freepool_agg.name)
                    LOG.info("removing host '%(host)s' from freepool "
                             "aggregate %(name)s",
                             {'host': host, 'name': freepool_agg.name})
                    try:
                        self.remove_computehost(freepool_agg.id, host)
                        removed_hosts.append(host)
                    except nova_exception.NotFound:
                        raise manager_exceptions.HostNotFound(host=host)

                    # When moving a host out of the freepool, we need to
                    # terminate preemptible instances before adding hosts to
                    # the reservation aggregate, which makes them available for
                    # scheduling.
                    #
                    # NOTE(priteau): Preemptibles should not be used with
                    # instance reservation yet.
                    self.terminate_preemptibles(host)

                LOG.info("adding host '%(host)s' to aggregate %(id)s",
                         {'host': host, 'id': agg.id})
                try:
                    self.nova.aggregates.add_host(agg.id, host)
                    added_hosts.append(host)
                except nova_exception.NotFound:
                    raise manager_exceptions.HostNotFound(host=host)
                except nova_exception.Conflict as e:
                    raise manager_exceptions.AggregateAlreadyHasHost(
                        pool=pool, host=host, nova_exception=str(e))
        except Exception as e:
            if added_hosts:
                LOG.warning('Removing hosts added to aggregate %s: %s',
                            agg.id, added_hosts)
                for host in added_hosts:
                    self.nova.aggregates.remove_host(agg.id, host)
            if removed_hosts:
                LOG.warning('Adding hosts back to freepool: %s', removed_hosts)
                for host in removed_hosts:
                    self.nova.aggregates.add_host(freepool_agg.id, host)
            raise e

        return self.get_aggregate_from_name_or_id(pool)

    def remove_all_computehosts(self, pool):
        """Remove all compute hosts attached to an aggregate."""

        hosts = self.get_computehosts(pool)
        self.remove_computehost(pool, hosts)

    def remove_computehost(self, pool, hosts):
        """Remove compute host(s) from an aggregate."""

        if not isinstance(hosts, list):
            hosts = [hosts]

        agg = self.get_aggregate_from_name_or_id(pool)

        try:
            freepool_agg = self.get(self.freepool_name)
        except manager_exceptions.AggregateNotFound:
            raise manager_exceptions.NoFreePool()

        hosts_failing_to_remove = []
        hosts_failing_to_add = []
        hosts_not_in_freepool = []
        for host in hosts:
            if freepool_agg.id == agg.id:
                if host not in freepool_agg.hosts:
                    hosts_not_in_freepool.append(host)
                    continue
            try:
                self.nova.aggregates.remove_host(agg.id, host)
            except nova_exception.ClientException:
                hosts_failing_to_remove.append(host)
            if freepool_agg.id != agg.id and host not in freepool_agg.hosts:
                # NOTE(sbauza) : We don't want to put again the host in
                # freepool if the requested pool is the freepool...
                try:
                    self.nova.aggregates.add_host(freepool_agg.id, host)
                except nova_exception.ClientException:
                    hosts_failing_to_add.append(host)

        if hosts_failing_to_remove:
            raise manager_exceptions.CantRemoveHost(
                host=hosts_failing_to_remove, pool=agg)
        if hosts_failing_to_add:
            raise manager_exceptions.CantAddHost(host=hosts_failing_to_add,
                                                 pool=freepool_agg)
        if hosts_not_in_freepool:
            raise manager_exceptions.HostNotInFreePool(
                host=hosts_not_in_freepool, freepool_name=freepool_agg.name)

    def add_project(self, pool, project_id):
        """Add a project to an aggregate."""

        metadata = {project_id: self.config.project_id_key}

        agg = self.get_aggregate_from_name_or_id(pool)

        return self.nova.aggregates.set_metadata(agg.id, metadata)

    def remove_project(self, pool, project_id):
        """Remove a project from an aggregate."""

        agg = self.get_aggregate_from_name_or_id(pool)

        metadata = {project_id: None}
        return self.nova.aggregates.set_metadata(agg.id, metadata)

    def terminate_preemptibles(self, host):
        """Terminate preemptible instances running on host"""
        for server in self.nova.servers.list(
                search_opts={"host": host, "all_tenants": 1}):
            try:
                LOG.info('Terminating preemptible instance %s (%s)',
                         server.name, server.id)
                self.nova.servers.delete(server=server)
            except nova_exception.NotFound:
                LOG.info('Could not find server %s, may have been deleted '
                         'concurrently.', server)
            except Exception as e:
                LOG.exception('Failed to delete %s: %s.', server, str(e))


class NovaInventory(NovaClientWrapper):

    def get_host_details(self, host):
        """Get Nova capabilities of a single host

        :param host: UUID, ID or name of nova-compute host
        :return: Dict of capabilities or raise HostNotFound
        """
        try:
            # NOTE(tetsuro): Only id (microversion < 2.53) or uuid
            # (microversion >= 2.53) is acceptable for the
            # `novaclient.hypervisors.get` argument. The invalid arguments
            # result in NotFound exception for microversion < 2.53 and
            # BadRequest exception for microversion >= 2.53
            hypervisor = self.nova.hypervisors.get(host)
        except (nova_exception.NotFound, nova_exception.BadRequest):
            # Name (not id or uuid) is given for the `host` parameter.
            try:
                hypervisors_list = self.nova.hypervisors.search(host)
            except nova_exception.NotFound:
                raise manager_exceptions.HostNotFound(host=host)
            if len(hypervisors_list) > 1:
                raise manager_exceptions.MultipleHostsFound(host=host)
            else:
                hypervisor_id = hypervisors_list[0].id
                # NOTE(sbauza): No need to catch the exception as we're sure
                #  that the hypervisor exists
                hypervisor = self.nova.hypervisors.get(hypervisor_id)

        az_name = ''
        if CONF.nova.az_aware:
            host_name = hypervisor.service['host']
            for zone in self.nova.availability_zones.list(detailed=True):
                if (zone.hosts and host_name in zone.hosts
                   and 'nova-compute' in zone.hosts[host_name]):
                    az_name = zone.zoneName

        try:
            # NOTE(tetsuro): compute API microversion 2.28 changes cpu_info
            # from string to object
            cpu_info = str(hypervisor.cpu_info)
            return {'id': hypervisor.id,
                    'availability_zone': az_name,
                    'hypervisor_hostname': hypervisor.hypervisor_hostname,
                    'service_name': hypervisor.service['host'],
                    'vcpus': hypervisor.vcpus,
                    'cpu_info': cpu_info,
                    'hypervisor_type': hypervisor.hypervisor_type,
                    'hypervisor_version': hypervisor.hypervisor_version,
                    'memory_mb': hypervisor.memory_mb,
                    'local_gb': hypervisor.local_gb}
        except AttributeError:
            raise manager_exceptions.InvalidHost(host=host)

    def get_servers_per_host(self, host):
        """List all servers of a nova-compute host

        :param host: Name (not UUID) of nova-compute host
        :return: Dict of servers or None
        """
        try:
            hypervisors_list = self.nova.hypervisors.search(host, servers=True)
        except nova_exception.NotFound:
            raise manager_exceptions.HostNotFound(host=host)
        if len(hypervisors_list) > 1:
            raise manager_exceptions.MultipleHostsFound(host=host)
        else:
            try:
                return hypervisors_list[0].servers
            except AttributeError:
                # NOTE(sbauza): nova.hypervisors.search(servers=True) returns
                #  a list of hosts without 'servers' attribute if no servers
                #  are running on that host
                return None



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import flask

from blazar.api.v1 import app as v1_app
from blazar.api.v1 import utils
from blazar import tests


class Error(Exception):
    def __init__(self, message=None, code=None):
        self.code = code
        super(Error, self).__init__(message)


class UtilsTestCase(tests.TestCase):
    def setUp(self):
        super(UtilsTestCase, self).setUp()
        app = v1_app.make_app()

        with app.test_request_context():
            self.utils = utils
            self.flask = flask

            self.rest = utils.Rest('v1_0', __name__)
            self.patch(self.rest, "_mroute")
            self.response = self.patch(self.flask, "Response")
            self.abort = self.patch(self.flask, "abort")
            self.request = self.patch(self.flask, "request")

            self.error = Error("message", "code")

    def test_get(self):
        self.rest.get('rule', status_code=200)
        self.rest._mroute.assert_called_once_with('GET', 'rule', 200,
                                                  query=False)

    def test_post(self):
        self.rest.post('rule', status_code=201)
        self.rest._mroute.assert_called_once_with('POST', 'rule', 201)

    def test_put(self):
        self.rest.put('rule', status_code=200)
        self.rest._mroute.assert_called_once_with('PUT', 'rule', 200)

    def test_delete(self):
        self.rest.delete('rule', status_code=204)
        self.rest._mroute.assert_called_once_with('DELETE', 'rule', 204)

    def test_route_ok(self):
        pass

    def test_route_fail(self):
        pass

    def test_render_wrong_result(self):
        self.utils.render(result=['a', 'a'], response_type='application/json',
                          status='LOL', kwargs={'a': 'b'})
        self.abort.assert_called_once_with(
            500, description="Non-dict and non-empty kwargs passed to render.")

    def test_render_wrong_resp_type(self):
        self.utils.render(result={}, response_type="not_app", status='LOL')
        self.abort.assert_called_once_with(
            400, description="Content type 'not_app' isn't supported")

    def test_render_ok(self):
        self.utils.render(result={}, response_type='application/json',
                          status='lol')
        self.response.assert_called_once_with(mimetype='application/json',
                                              status='lol', response=b'{}')

    def test_request_data_data(self):
        self.request.parsed_data = "data"
        self.assertEqual("data", self.utils.request_data())

    def test_request_data_file(self):
        self.request.file_upload = True
        self.request.data = 'foo'
        del self.request.parsed_data
        flask.request.content_length = 1
        self.assertEqual('foo', self.utils.request_data())

    def test_request_data_length(self):
        del self.request.parsed_data
        self.request.content_length = 0
        self.utils.request_data()
        self.assertEqual({}, self.utils.request_data())

    def test_get_request_args(self):
        self.flask.request.args = 'foo'
        self.assertEqual('foo', self.utils.get_request_args())

    def test_abort_and_log(self):
        self.utils.abort_and_log(400, "Funny error")
        self.abort.assert_called_once_with(400, description="Funny error")

    def test_render_error_message(self):
        render = self.patch(self.utils, 'render')
        self.utils.render_error_message(404, 'NOT FOUND', 'not_found')
        render.assert_called_once_with({'error_name': 'not_found',
                                        'error_message': 'NOT FOUND',
                                        'error_code': 404})

    def test_internal_error_501(self):
        error_message = self.patch(self.utils, 'render_error_message')
        self.utils.internal_error(501, "Funny error")
        error_message.assert_called_once_with(
            501, "Funny error", "NOT_IMPLEMENTED_ERROR")

    def test_internal_error_various(self):
        error_message = self.patch(self.utils, 'render_error_message')
        self.utils.internal_error(404, "Funny error")
        error_message.assert_called_once_with(
            404, "Funny error", "INTERNAL_SERVER_ERROR")

    def test_bad_request(self):
        error_message = self.patch(self.utils, 'render_error_message')
        self.utils.bad_request(self.error)
        error_message.assert_called_once_with('code', 'message', 'code')

    def test_bad_request_with_default_errorcode(self):
        error_message = self.patch(self.utils, 'render_error_message')
        error = Error("message")
        self.utils.bad_request(error)
        error_message.assert_called_once_with(400, 'message', 400)

    def test_not_found(self):
        error_message = self.patch(self.utils, 'render_error_message')
        self.utils.not_found(self.error)
        error_message.assert_called_once_with('code', 'message', 'code')

    def test_not_found_with_default_errorcode(self):
        error_message = self.patch(self.utils, 'render_error_message')
        error = Error("message")
        self.utils.not_found(error)
        error_message.assert_called_once_with(404, 'message', 404)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Simple middleware for request logging."""

from oslo_log import log as logging


LOG = logging.getLogger(__name__)


class RequestLog(object):
    """Middleware to write a simple request log to.

    Borrowed from Paste Translogger
    """

    format = ('%(REMOTE_ADDR)s "%(REQUEST_METHOD)s %(REQUEST_URI)s" '
              'status: %(status)s len: %(bytes)s ')

    def __init__(self, application):
        self.application = application

    def __call__(self, environ, start_response):
        LOG.debug('Starting request: %s "%s %s"',
                  environ['REMOTE_ADDR'], environ['REQUEST_METHOD'],
                  self._get_uri(environ))
        # Set the accept header if it is not otherwise set or is '*/*'. This
        # ensures that error responses will be in JSON.
        accept = environ.get('HTTP_ACCEPT')
        if accept:
            environ['HTTP_ACCEPT'] = 'application/json'
        if LOG.isEnabledFor(logging.INFO):
            return self._log_app(environ, start_response)

    @staticmethod
    def _get_uri(environ):
        req_uri = (environ.get('SCRIPT_NAME', '')
                   + environ.get('PATH_INFO', ''))
        if environ.get('QUERY_STRING'):
            req_uri += '?' + environ['QUERY_STRING']
        return req_uri

    def _log_app(self, environ, start_response):
        req_uri = self._get_uri(environ)

        def replacement_start_response(status, headers, exc_info=None):
            """We need to gaze at the content-length, if set to write log info.

            """
            size = None
            for name, value in headers:
                if name.lower() == 'content-length':
                    size = value
            self.write_log(environ, req_uri, status, size)
            return start_response(status, headers, exc_info)

        return self.application(environ, replacement_start_response)

    def write_log(self, environ, req_uri, status, size):
        """Write the log info out in a formatted form to ``LOG.info``.

        """
        if size is None:
            size = '-'
        log_format = {
            'REMOTE_ADDR': environ.get('REMOTE_ADDR', '-'),
            'REQUEST_METHOD': environ['REQUEST_METHOD'],
            'REQUEST_URI': req_uri,
            'status': status.split(None, 1)[0],
            'bytes': size
        }
        LOG.info(self.format, log_format)



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.api.v1.leases import service as service_api
from blazar import tests


class RPCApiTestCase(tests.TestCase):
    def setUp(self):
        super(RPCApiTestCase, self).setUp()
        self.s_api = service_api

        self.fake_list = []
        self.fake_lease = {}

        self.patch(self.s_api.API, "get_leases").return_value = self.fake_list
        self.patch(self.s_api.API, "create_lease").return_value = True
        self.patch(self.s_api.API, "get_lease").return_value = self.fake_lease
        self.patch(self.s_api.API, "update_lease").return_value = True
        self.patch(self.s_api.API, "delete_lease").return_value = True
        self.patch(self.s_api.API, "get_plugins").return_value(self.fake_lease)

    def test_get_lease(self):
        pass

    def test_create_lease(self):
        pass

    def test_update_lease(self):
        pass

    def test_delete_lease(self):
        pass

    def get_plugins(self):
        pass



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
#
# Copyright © 2013  Julien Danjou <julien@danjou.info>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools

from oslo_config import cfg
from oslo_log import log as logging
from oslo_service import service

from blazar import context
from blazar import rpc

LOG = logging.getLogger(__name__)


class RPCClient(object):
    def __init__(self, target):
        super(RPCClient, self).__init__()
        rpc.init()
        self._client = rpc.get_client(target)

    def cast(self, name, **kwargs):
        ctx = context.current()
        self._client.cast(ctx.to_dict(), name, **kwargs)

    def call(self, name, **kwargs):
        ctx = context.current()
        return self._client.call(ctx.to_dict(), name, **kwargs)


class RPCServer(service.Service):
    def __init__(self, target):
        super(RPCServer, self).__init__()
        rpc.init()
        self._server = rpc.get_server(
            target,
            endpoints=[ContextEndpointHandler(self, target)]
        )

    def start(self):
        super(RPCServer, self).start()
        self.tg.add_thread(self._server.start)

    def wait(self):
        super(RPCServer, self).wait()
        self._server.wait()

    def stop(self):
        super(RPCServer, self).stop()
        self._server.stop()


class ContextEndpointHandler(object):
    def __init__(self, endpoint, target):
        self.__endpoint = endpoint
        self.target = target

    def __getattr__(self, name):
        try:
            method = getattr(self.__endpoint, name)

            def run_method(__ctx, **kwargs):
                with context.BlazarContext.from_dict(__ctx):
                    return method(**kwargs)

            return run_method
        except AttributeError:
            LOG.error("No %(method)s method found implemented in "
                      "%(class)s class",
                      {'method': name, 'class': self.__endpoint})


def with_empty_context(func):
    @functools.wraps(func)
    def decorator(*args, **kwargs):
        with context.BlazarContext():
            return func(*args, **kwargs)

    return decorator


def prepare_service(argv=[]):
    logging.setup(cfg.CONF, 'blazar')



# Copyright 2017 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Allow duplicate lease names

Revision ID: 6bfd1c23aa18
Revises: ba75b766b64e
Create Date: 2017-08-16 08:49:15.307072

"""

# revision identifiers, used by Alembic.
revision = '6bfd1c23aa18'
down_revision = 'ba75b766b64e'

from alembic import op
from sqlalchemy.engine import reflection


def upgrade():
    inspector = reflection.Inspector.from_engine(op.get_bind())
    unique_constraints = inspector.get_unique_constraints('leases')
    for constraint in unique_constraints:
        if constraint['column_names'] == ['name']:
            op.drop_constraint(constraint['name'], 'leases', type_='unique')
            break


def downgrade():
    op.create_unique_constraint('uniq_leases0name', 'leases', ['name'])



# Copyright (c) 2013 Bull.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.api.v1.oshosts import service
from blazar.api.v1 import utils as api_utils
from blazar.api.v1 import validation
from blazar import utils


def get_rest():
    """Return Rest app"""
    return rest


rest = api_utils.Rest('host_v1_0', __name__, url_prefix='/v1/os-hosts')
_api = utils.LazyProxy(service.API)


# Computehosts operations

@rest.get('', query=True)
def computehosts_list(req, query=None):
    """List all existing computehosts."""
    return api_utils.render(hosts=_api.get_computehosts(query))


@rest.post('')
def computehosts_create(req, data):
    """Create new computehost."""
    return api_utils.render(host=_api.create_computehost(data))


@rest.get('/<host_id>')
@validation.check_exists(_api.get_computehost, host_id='host_id')
def computehosts_get(req, host_id):
    """Get computehost by its ID."""
    return api_utils.render(host=_api.get_computehost(host_id))


@rest.put('/<host_id>')
@validation.check_exists(_api.get_computehost, host_id='host_id')
def computehosts_update(req, host_id, data):
    """Update computehost. Only name changing may be proceeded."""
    if len(data) == 0:
        return api_utils.internal_error(status_code=400,
                                        descr="No data to update")
    else:
        return api_utils.render(host=_api.update_computehost(host_id, data))


@rest.delete('/<host_id>')
@validation.check_exists(_api.get_computehost, host_id='host_id')
def computehosts_delete(req, host_id):
    """Delete specified computehost."""
    _api.delete_computehost(host_id)
    return api_utils.render()


@rest.get('/allocations', query=True)
def allocations_list(req, query):
    """List all allocations on all computehosts."""
    return api_utils.render(allocations=_api.list_allocations(query))


@rest.get('/<host_id>/allocation', query=True)
@validation.check_exists(_api.get_computehost, host_id='host_id')
def allocations_get(req, host_id, query):
    """List all allocations on a specific host."""
    return api_utils.render(allocation=_api.get_allocations(host_id, query))


@rest.get('/properties', query=True)
def resource_properties_list(req, query=None):
    """List computehost resource properties."""
    return api_utils.render(
        resource_properties=_api.list_resource_properties(query))


@rest.patch('/properties/<property_name>')
def resource_property_update(req, property_name, data):
    """Update a computehost resource property."""
    return api_utils.render(
        resource_property=_api.update_resource_property(property_name, data))



# Copyright 2024 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add compute host inventory and traits

Revision ID: 553383923ca0
Revises: 02e2f2186d98
Create Date: 2024-04-29 17:40:05.148493

"""

# revision identifiers, used by Alembic.
revision = '553383923ca0'
down_revision = '02e2f2186d98'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'computehost_resource_inventory',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('computehost_id', sa.String(length=36), nullable=True),
        sa.Column('resource_class', sa.String(length=255), nullable=False),
        sa.Column('total', sa.Integer(), nullable=False),
        sa.Column('reserved', sa.Integer(), nullable=False),
        sa.Column('min_unit', sa.Integer(), nullable=False),
        sa.Column('max_unit', sa.Integer(), nullable=False),
        sa.Column('step_size', sa.Integer(), nullable=False),
        sa.Column('allocation_ratio', sa.Float(), nullable=False),
        sa.ForeignKeyConstraint(['computehost_id'], ['computehosts.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'computehost_trait',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('computehost_id', sa.String(length=36), nullable=True),
        sa.Column('trait', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['computehost_id'], ['computehosts.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('computehost_trait')
    op.drop_table('computehost_resource_inventory')



# Copyright 2014 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add trust_id to ComputeHost

Revision ID: 1fd6c2eded89
Revises: 0_1
Create Date: 2014-03-28 01:12:02.735519

"""

# revision identifiers, used by Alembic.
revision = '1fd6c2eded89'
down_revision = '23d6240b51b2'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('computehosts',
                  sa.Column('trust_id',
                            sa.String(length=36),
                            nullable=True))

    if op.get_bind().engine.name != 'sqlite':
        # I need to do it in this way because Postgress fails
        # if I use SQLAlchemy
        connection = op.get_bind()
        connection.execute(sa.text("UPDATE computehosts SET trust_id = ''"))

        op.alter_column('computehosts', 'trust_id',
                        existing_type=sa.String(length=36), nullable=False)


def downgrade():
    engine = op.get_bind().engine
    if engine.name == 'sqlite':
        # Only for testing purposes with sqlite
        op.execute('CREATE TABLE tmp_computehosts as SELECT id, '
                   'vcpus, cpu_info, hypervisor_type, '
                   'hypervisor_version, hypervisor_hostname, service_name, '
                   'memory_mb, local_gb, status '
                   'FROM computehosts')
        op.execute('DROP TABLE computehosts')
        op.execute('ALTER TABLE tmp_computehosts RENAME TO computehosts')
        return

    op.drop_column('computehosts', 'trust_id')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from blazar.api.v1.leases import service as service_api
from blazar.api.v1 import utils as api_utils
from blazar.api.v1 import validation as validation_api
from blazar import exceptions
from blazar import tests


class ValidationTestCase(tests.TestCase):
    def setUp(self):
        super(ValidationTestCase, self).setUp()

        self.s_api = service_api
        self.u_api = api_utils
        self.v_api = validation_api
        self.exc = exceptions

        self.patch(self.u_api, 'render')
        self.not_found = self.patch(self.u_api, 'not_found')

        self.fake_id = 1

    def test_check_true(self):
        fake_get = self.patch(self.s_api.API, 'get_lease').return_value = True

        @self.v_api.check_exists(fake_get, self.fake_id)
        def trap(fake_id):
            self.u_api.render(lease_id=self.fake_id)
            fake_get.assert_called_once_with()

    def test_check_false(self):
        fake_get = self.patch(
            self.s_api.API, 'get_lease').side_effect = self.exc.NotFound()

        @self.v_api.check_exists(fake_get, self.fake_id)
        def trap(fake_id):
            self.u_api.render(lease_id=self.fake_id)
            self.not_found.assert_called_once_with()



# Copyright 2018 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""no affinity instance reservation

Revision ID: 9593f3656974
Revises: 35b314cd39ee
Create Date: 2018-12-26 15:55:29.950250

"""

# revision identifiers, used by Alembic.
revision = '9593f3656974'
down_revision = '35b314cd39ee'

from alembic import op
from sqlalchemy.dialects import mysql


def upgrade():
    op.alter_column('instance_reservations', 'affinity',
                    existing_type=mysql.TINYINT(display_width=1),
                    nullable=True)


def downgrade():
    # TODO(tetsuro): Since the Kilo release, OpenStack doesn't support
    # database downgrades, we should delete this, update the related
    # Blazar documents, and issue a reno in a followup.
    op.alter_column('instance_reservations', 'affinity',
                    existing_type=mysql.TINYINT(display_width=1),
                    nullable=False)



# Copyright (c) 2017 NTT Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_serialization import jsonutils

from blazar.api.v1 import app as v1_app
from blazar.api.v2 import app as v2_app


class VersionSelectorApplication(object):
    """Maps WSGI versioned apps and defines default WSGI app."""

    def __init__(self):
        self._status = ''
        self._response_headers = []
        self.v1 = v1_app.make_app()
        self.v2 = v2_app.make_app()

    def _append_versions_from_app(self, versions, app, environ):
        tmp_versions = app(environ, self.internal_start_response)
        if self._status.startswith("300"):
            # In case of v1, app returns ClosingIterator generator object,
            # whereas in case of v2, it returns list.
            # So convert it to iterator to get the versions.
            app_iter = iter(tmp_versions)
            tmp_versions = jsonutils.loads(next(app_iter))
            versions['versions'].extend(tmp_versions['versions'])
        return tmp_versions

    def internal_start_response(self, status, response_headers, exc_info=None):
        self._status = status
        self._response_headers = response_headers

    def __call__(self, environ, start_response):
        self._status = ''
        self._response_headers = []

        if environ['PATH_INFO'] == '/' or environ['PATH_INFO'] == '/versions':
            versions = {'versions': []}
            tmp_versions = self._append_versions_from_app(versions, self.v1,
                                                          environ)
            # Both v1 and v2 apps run auth_token middleware. So simply
            # validate token for v1. If it fails no need to call v2 app.
            if self._status.startswith("401"):
                start_response(self._status,
                               [("Content-Type", "application/json")])
                return tmp_versions
            self._append_versions_from_app(versions, self.v2,
                                           environ)
            if len(versions['versions']):
                start_response("300 Multiple Choices",
                               [("Content-Type", "application/json")])
                return [jsonutils.dump_as_bytes(versions)]
            else:
                start_response("204 No Content", [])
                return []
        else:
            if environ['PATH_INFO'].startswith('/v1'):
                return self.v1(environ, start_response)
            return self.v2(environ, start_response)



# Copyright (c) 2019 NTT.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
from unittest import mock

from oslo_config import cfg

from blazar import context
from blazar.db import api as db_api
from blazar.db import utils as db_utils
from blazar.manager import exceptions as mgr_exceptions
from blazar.plugins import floatingips as plugin
from blazar.plugins.floatingips import floatingip_plugin
from blazar import tests
from blazar.utils.openstack import exceptions as opst_exceptions
from blazar.utils.openstack import neutron


CONF = cfg.CONF


class FloatingIpPluginTest(tests.TestCase):

    def setUp(self):
        super(FloatingIpPluginTest, self).setUp()

        self.cfg = cfg

        # Make sure we clean up any override which could impact other tests
        self.addCleanup(self.cfg.CONF.reset)

        self.db_api = db_api
        self.db_utils = db_utils
        self.fip_pool = self.patch(neutron, 'FloatingIPPool')

    def test_create_floatingip(self):
        m = mock.MagicMock()
        m.fetch_subnet.return_value = {'id': 'subnet-id'}
        self.fip_pool.return_value = m
        fip_row = {
            'id': 'fip-id',
            'network_id': 'net-id',
            'subnet_id': 'subnet-id',
            'floating_ip_address': '172.24.4.100',
            'reservable': True
        }
        patch_fip_create = self.patch(db_api, 'floatingip_create')
        patch_fip_create.return_value = fip_row

        data = {
            'floating_ip_address': '172.24.4.100',
            'floating_network_id': 'net-id'
        }
        expected = fip_row

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        ret = fip_plugin.create_floatingip(data)

        self.assertDictEqual(expected, ret)
        m.fetch_subnet.assert_called_once_with('172.24.4.100')
        patch_fip_create.assert_called_once_with({
            'floating_network_id': 'net-id',
            'subnet_id': 'subnet-id',
            'floating_ip_address': '172.24.4.100'})

    def test_create_floatingip_with_invalid_ip(self):
        m = mock.MagicMock()
        m.fetch_subnet.side_effect = opst_exceptions.NeutronUsesFloatingIP()
        self.fip_pool.return_value = m

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        self.assertRaises(opst_exceptions.NeutronUsesFloatingIP,
                          fip_plugin.create_floatingip,
                          {'floating_ip_address': 'invalid-ip',
                           'floating_network_id': 'id'})

    def test_get_floatingip(self):
        fip_row = {
            'id': 'fip-id',
            'network_id': 'net-id',
            'subnet_id': 'subnet-id',
            'floating_ip_address': '172.24.4.100',
            'reservable': True
        }
        patch_fip_get = self.patch(db_api, 'floatingip_get')
        patch_fip_get.return_value = fip_row

        expected = fip_row

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        ret = fip_plugin.get_floatingip('fip-id')

        self.assertDictEqual(expected, ret)
        patch_fip_get.assert_called_once_with('fip-id')

    def test_get_floatingip_with_no_exist(self):
        patch_fip_get = self.patch(db_api, 'floatingip_get')
        patch_fip_get.return_value = None

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        self.assertRaises(mgr_exceptions.FloatingIPNotFound,
                          fip_plugin.get_floatingip, 'fip-id')

        patch_fip_get.assert_called_once_with('fip-id')

    def test_get_list_floatingips(self):
        fip_rows = [{
            'id': 'fip-id',
            'network_id': 'net-id',
            'subnet_id': 'subnet-id',
            'floating_ip_address': '172.24.4.100',
            'reservable': True
        }]
        patch_fip_list = self.patch(db_api, 'floatingip_list')
        patch_fip_list.return_value = fip_rows

        expected = fip_rows

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        ret = fip_plugin.list_floatingip()

        self.assertListEqual(expected, ret)
        patch_fip_list.assert_called_once_with()

    def test_delete_floatingip(self):
        fip_row = {
            'id': 'fip-id',
            'network_id': 'net-id',
            'subnet_id': 'subnet-id',
            'floating_ip_address': '172.24.4.100',
            'reservable': True
        }
        patch_fip_get = self.patch(db_api, 'floatingip_get')
        patch_fip_get.return_value = fip_row
        patch_fip_alloc = self.patch(db_api,
                                     'fip_allocation_get_all_by_values')
        patch_fip_alloc.return_value = []
        patch_fip_destroy = self.patch(db_api, 'floatingip_destroy')

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        fip_plugin.delete_floatingip('fip-id')

        patch_fip_get.assert_called_once_with('fip-id')
        patch_fip_alloc.assert_called_once_with(floatingip_id='fip-id')
        patch_fip_destroy.assert_called_once_with('fip-id')

    def test_delete_floatingip_with_reservations(self):
        fip_row = {
            'id': 'fip-id',
            'network_id': 'net-id',
            'subnet_id': 'subnet-id',
            'floating_ip_address': '172.24.4.100',
            'reservable': True
        }
        patch_fip_get = self.patch(db_api, 'floatingip_get')
        patch_fip_get.return_value = fip_row
        patch_fip_alloc = self.patch(db_api,
                                     'fip_allocation_get_all_by_values')
        patch_fip_alloc.return_value = [
            {
                'id': 'alloc-id1',
                'floatingip_id': 'fip-id',
                'reservation_id': 'reservations-id1'
            }
        ]
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        self.assertRaises(mgr_exceptions.CantDeleteFloatingIP,
                          fip_plugin.delete_floatingip,
                          'fip-id1')

    def test_delete_floatingip_with_no_exist(self):
        patch_fip_get = self.patch(db_api, 'floatingip_get')
        patch_fip_get.return_value = None

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        self.assertRaises(mgr_exceptions.FloatingIPNotFound,
                          fip_plugin.delete_floatingip,
                          'non-exists-id')

    def test_create_reservation_fips_available(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        values = {
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
            'resource_type': plugin.RESOURCE_TYPE,
            'amount': 2
        }
        matching_fips = self.patch(fip_plugin, '_matching_fips')
        matching_fips.return_value = ['fip1', 'fip2']
        fip_reservation_create = self.patch(self.db_api,
                                            'fip_reservation_create')
        fip_allocation_create = self.patch(
            self.db_api, 'fip_allocation_create')
        fip_plugin.reserve_resource(
            '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            values)
        fip_values = {
            'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'amount': 2
        }
        fip_reservation_create.assert_called_once_with(fip_values)
        calls = [
            mock.call(
                {'floatingip_id': 'fip1',
                 'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
                 }),
            mock.call(
                {'floatingip_id': 'fip2',
                 'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
                 }),
        ]
        fip_allocation_create.assert_has_calls(calls)

    def test_create_reservation_fips_with_required(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        values = {
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
            'resource_type': plugin.RESOURCE_TYPE,
            'amount': 2,
            'required_floatingips': ['172.24.4.100']
        }
        matching_fips = self.patch(fip_plugin, '_matching_fips')
        matching_fips.return_value = ['fip1', 'fip2']
        fip_reservation_create = self.patch(self.db_api,
                                            'fip_reservation_create')
        fip_reservation_create.return_value = {'id': 'fip_resv_id1'}
        fip_allocation_create = self.patch(
            self.db_api, 'fip_allocation_create')
        required_addr_create = self.patch(self.db_api, 'required_fip_create')
        fip_plugin.reserve_resource(
            '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            values)
        fip_values = {
            'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'amount': 2
        }
        fip_reservation_create.assert_called_once_with(fip_values)
        required_addr_create.assert_called_once_with(
            {
                'address': '172.24.4.100',
                'floatingip_reservation_id': 'fip_resv_id1'
            })
        calls = [
            mock.call(
                {'floatingip_id': 'fip1',
                 'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
                 }),
            mock.call(
                {'floatingip_id': 'fip2',
                 'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
                 }),
        ]
        fip_allocation_create.assert_has_calls(calls)

    def test_create_reservation_with_missing_param_network(self):
        values = {
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'amount': 2,
            'start_date': datetime.datetime(2017, 3, 1, 20, 0),
            'end_date': datetime.datetime(2017, 3, 2, 20, 0),
            'resource_type': plugin.RESOURCE_TYPE,
        }
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        self.assertRaises(
            mgr_exceptions.MissingParameter,
            fip_plugin.reserve_resource,
            '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            values)

    def test_create_reservation_with_invalid_fip(self):
        values = {
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'network_id': 'a37a14f3-e3eb-4fe2-9e36-082b67f12ea0',
            'amount': 2,
            'required_floatingips': ['aaa.aaa.aaa.aaa'],
            'start_date': datetime.datetime(2017, 3, 1, 20, 0),
            'end_date': datetime.datetime(2017, 3, 2, 20, 0),
            'resource_type': plugin.RESOURCE_TYPE,
        }
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        self.assertRaises(
            mgr_exceptions.InvalidIPFormat,
            fip_plugin.reserve_resource,
            '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            values)

    def test_create_reservation_required_bigger_than_amount(self):
        values = {
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'network_id': 'a37a14f3-e3eb-4fe2-9e36-082b67f12ea0',
            'amount': 1,
            'required_floatingips': ['172.24.4.100', '172.24.4.101'],
            'start_date': datetime.datetime(2017, 3, 1, 20, 0),
            'end_date': datetime.datetime(2017, 3, 2, 20, 0),
            'resource_type': plugin.RESOURCE_TYPE,
        }
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        self.assertRaises(
            mgr_exceptions.TooLongFloatingIPs,
            fip_plugin.reserve_resource,
            '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            values)

    def test_update_pending_reservation_increase_amount_fips_available(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        values = {
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
            'resource_type': plugin.RESOURCE_TYPE,
            'amount': 2,
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = {
            'id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
        }
        reservation_get = self.patch(self.db_api, 'reservation_get')
        reservation_get.return_value = {
            'id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'resource_id': 'fip-reservation-id-1',
            'resource_type': 'virtual:floatingip',
            'status': 'pending',
        }
        fip_allocation_get_all_by_values = self.patch(
            self.db_api, 'fip_allocation_get_all_by_values'
        )
        fip_allocation_get_all_by_values.return_value = [{
            'floatingip_id': 'fip1'
        }]
        fip_reservation_get = self.patch(self.db_api, 'fip_reservation_get')
        fip_reservation_get.return_value = {
            'id': 'fip_resv_id1',
            'amount': 1,
            'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'required_floatingips': [],
        }
        matching_fips = self.patch(fip_plugin, '_matching_fips')
        matching_fips.return_value = ['fip2']
        fip_reservation_update = self.patch(self.db_api,
                                            'fip_reservation_update')
        fip_allocation_create = self.patch(
            self.db_api, 'fip_allocation_create')
        fip_plugin.update_reservation(
            '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            values)
        fip_reservation_update.assert_called_once_with(
            'fip_resv_id1', {'amount': 2})
        calls = [
            mock.call(
                {'floatingip_id': 'fip2',
                 'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
                 })
        ]
        fip_allocation_create.assert_has_calls(calls)

    def test_update_active_reservation_increase_amount_fips_available(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        values = {
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
            'resource_type': plugin.RESOURCE_TYPE,
            'amount': 2,
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = {
            'id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
        }
        reservation_get = self.patch(self.db_api, 'reservation_get')
        reservation_get.return_value = {
            'id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'resource_id': 'fip-reservation-id-1',
            'resource_type': 'virtual:floatingip',
            'status': 'active',
        }
        fip_allocation_get_all_by_values = self.patch(
            self.db_api, 'fip_allocation_get_all_by_values'
        )
        fip_allocation_get_all_by_values.return_value = [{
            'floatingip_id': 'fip1'
        }]
        fip_reservation_get = self.patch(self.db_api, 'fip_reservation_get')
        fip_reservation_get.return_value = {
            'id': 'fip_resv_id1',
            'amount': 1,
            'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'required_floatingips': [],
        }
        matching_fips = self.patch(fip_plugin, '_matching_fips')
        matching_fips.return_value = ['fip2']
        fip_reservation_update = self.patch(self.db_api,
                                            'fip_reservation_update')
        fip_get = self.patch(self.db_api, 'floatingip_get')
        fip_get.return_value = {
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'subnet_id': 'subnet-id',
            'floating_ip_address': '172.2.24.100'
        }
        self.set_context(context.BlazarContext(project_id='fake-project-id'))
        m = mock.MagicMock()
        self.fip_pool.return_value = m
        fip_allocation_create = self.patch(
            self.db_api, 'fip_allocation_create')
        fip_allocation_destroy = self.patch(
            self.db_api, 'fip_allocation_destroy')
        fip_plugin.update_reservation(
            '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            values)
        fip_reservation_update.assert_called_once_with(
            'fip_resv_id1', {'amount': 2})
        self.fip_pool.assert_called_once_with(
            'f548089e-fb3e-4013-a043-c5ed809c7a67')
        m.create_reserved_floatingip.assert_called_once_with(
            'subnet-id',
            '172.2.24.100',
            'fake-project-id',
            '441c1476-9f8f-4700-9f30-cd9b6fef3509')
        calls = [
            mock.call(
                {'floatingip_id': 'fip2',
                 'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
                 })
        ]
        fip_allocation_create.assert_has_calls(calls)
        self.assertFalse(fip_allocation_destroy.called)

    def test_update_active_reservation_fip_creation_failure(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        values = {
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
            'resource_type': plugin.RESOURCE_TYPE,
            'amount': 3,
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = {
            'id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
        }
        reservation_get = self.patch(self.db_api, 'reservation_get')
        reservation_get.return_value = {
            'id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'resource_id': 'fip-reservation-id-1',
            'resource_type': 'virtual:floatingip',
            'status': 'active',
        }
        fip_allocation_get_all_by_values = self.patch(
            self.db_api, 'fip_allocation_get_all_by_values'
        )
        fip_allocation_get_all_by_values.return_value = [{
            'floatingip_id': 'fip1'
        }]
        fip_reservation_get = self.patch(self.db_api, 'fip_reservation_get')
        fip_reservation_get.return_value = {
            'id': 'fip_resv_id1',
            'amount': 1,
            'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'required_floatingips': [],
        }
        matching_fips = self.patch(fip_plugin, '_matching_fips')
        matching_fips.return_value = ['fip2', 'fip3']
        fip_get = self.patch(self.db_api, 'floatingip_get')
        fip_get.side_effect = (
            {
                'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
                'subnet_id': 'subnet-id',
                'floating_ip_address': '172.2.24.100'
            },
            {
                'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
                'subnet_id': 'subnet-id',
                'floating_ip_address': '172.2.24.101'
            }
        )
        self.set_context(context.BlazarContext(project_id='fake-project-id'))
        m = mock.MagicMock()
        m.create_reserved_floatingip.side_effect = (None, Exception())
        self.fip_pool.return_value = m
        fip_allocation_create = self.patch(
            self.db_api, 'fip_allocation_create')
        fip_allocation_destroy = self.patch(
            self.db_api, 'fip_allocation_destroy')
        self.assertRaises(mgr_exceptions.NeutronClientError,
                          fip_plugin.update_reservation,
                          '441c1476-9f8f-4700-9f30-cd9b6fef3509', values)
        self.fip_pool.assert_called_once_with(
            'f548089e-fb3e-4013-a043-c5ed809c7a67')
        calls = [
            mock.call('subnet-id',
                      '172.2.24.100',
                      'fake-project-id',
                      '441c1476-9f8f-4700-9f30-cd9b6fef3509'),
            mock.call('subnet-id',
                      '172.2.24.101',
                      'fake-project-id',
                      '441c1476-9f8f-4700-9f30-cd9b6fef3509'),
        ]
        m.create_reserved_floatingip.assert_has_calls(calls)
        self.assertFalse(fip_allocation_create.called)
        self.assertFalse(fip_allocation_destroy.called)

    def test_update_reservation_increase_amount_fips_unavailable(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        values = {
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
            'resource_type': plugin.RESOURCE_TYPE,
            'amount': 2,
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = {
            'id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
        }
        reservation_get = self.patch(self.db_api, 'reservation_get')
        reservation_get.return_value = {
            'id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'resource_id': 'fip-reservation-id-1',
            'resource_type': 'virtual:floatingip',
            'status': 'pending',
        }
        fip_allocation_get_all_by_values = self.patch(
            self.db_api, 'fip_allocation_get_all_by_values'
        )
        fip_allocation_get_all_by_values.return_value = [{
            'floatingip_id': 'fip1'
        }]
        fip_reservation_get = self.patch(self.db_api, 'fip_reservation_get')
        fip_reservation_get.return_value = {
            'id': 'fip_resv_id1',
            'amount': 1,
            'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'required_floatingips': [],
        }
        matching_fips = self.patch(fip_plugin, '_matching_fips')
        matching_fips.return_value = []
        self.assertRaises(mgr_exceptions.NotEnoughFloatingIPAvailable,
                          fip_plugin.update_reservation,
                          '441c1476-9f8f-4700-9f30-cd9b6fef3509', values)

    def test_update_reservation_decrease_amount(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        values = {
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
            'resource_type': plugin.RESOURCE_TYPE,
            'amount': 1,
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = {
            'id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
        }
        reservation_get = self.patch(self.db_api, 'reservation_get')
        reservation_get.return_value = {
            'id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'resource_id': 'fip-reservation-id-1',
            'resource_type': 'virtual:floatingip',
            'status': 'pending',
        }
        fip_allocation_get_all_by_values = self.patch(
            self.db_api, 'fip_allocation_get_all_by_values'
        )
        fip_allocation_get_all_by_values.return_value = [
            {'id': 'fip_alloc_1', 'floatingip_id': 'fip1'},
            {'id': 'fip_alloc_2', 'floatingip_id': 'fip2'},
        ]
        fip_reservation_get = self.patch(self.db_api, 'fip_reservation_get')
        fip_reservation_get.return_value = {
            'id': 'fip_resv_id1',
            'amount': 2,
            'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'required_floatingips': [],
        }
        fip_allocation_destroy = self.patch(self.db_api,
                                            'fip_allocation_destroy')
        fip_reservation_update = self.patch(self.db_api,
                                            'fip_reservation_update')
        fip_plugin.update_reservation(
            '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            values)
        fip_reservation_update.assert_called_once_with(
            'fip_resv_id1', {'amount': 1})
        calls = [
            mock.call('fip_alloc_1')
        ]
        fip_allocation_destroy.assert_has_calls(calls)

    def test_update_reservation_remove_required_fips(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        values = {
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
            'resource_type': plugin.RESOURCE_TYPE,
            'required_floatingips': [],
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = {
            'id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
        }
        reservation_get = self.patch(self.db_api, 'reservation_get')
        reservation_get.return_value = {
            'id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'resource_id': 'fip-reservation-id-1',
            'resource_type': 'virtual:floatingip',
            'status': 'pending',
        }
        fip_allocation_get_all_by_values = self.patch(
            self.db_api, 'fip_allocation_get_all_by_values'
        )
        fip_allocation_get_all_by_values.return_value = [{
            'floatingip_id': 'fip1'
        }]
        fip_reservation_get = self.patch(self.db_api, 'fip_reservation_get')
        fip_reservation_get.return_value = {
            'id': 'fip_resv_id1',
            'amount': 1,
            'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'required_floatingips': ['172.24.4.100']
        }
        required_fip_destroy_by_fip_reservation_id = self.patch(
            self.db_api, 'required_fip_destroy_by_fip_reservation_id')
        fip_plugin.update_reservation(
            '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            values)
        calls = [mock.call('fip_resv_id1')]
        required_fip_destroy_by_fip_reservation_id.assert_has_calls(calls)

    def test_update_reservation_change_required_fips(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        values = {
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
            'resource_type': plugin.RESOURCE_TYPE,
            'required_floatingips': ['172.24.4.101'],
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = {
            'id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
        }
        reservation_get = self.patch(self.db_api, 'reservation_get')
        reservation_get.return_value = {
            'id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'resource_id': 'fip-reservation-id-1',
            'resource_type': 'virtual:floatingip',
            'status': 'pending',
        }
        fip_reservation_get = self.patch(self.db_api, 'fip_reservation_get')
        fip_reservation_get.return_value = {
            'id': 'fip_resv_id1',
            'amount': 1,
            'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
            'required_floatingips': ['172.24.4.100']
        }
        self.assertRaises(mgr_exceptions.CantUpdateFloatingIPReservation,
                          fip_plugin.update_reservation,
                          '441c1476-9f8f-4700-9f30-cd9b6fef3509', values)

    def test_update_reservation_change_network_id(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        values = {
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
            'resource_type': plugin.RESOURCE_TYPE,
            'network_id': 'new-network-id',
        }
        lease_get = self.patch(self.db_api, 'lease_get')
        lease_get.return_value = {
            'id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'start_date': datetime.datetime(2013, 12, 19, 20, 0),
            'end_date': datetime.datetime(2013, 12, 19, 21, 0),
        }
        reservation_get = self.patch(self.db_api, 'reservation_get')
        reservation_get.return_value = {
            'id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'lease_id': '018c1b43-e69e-4aef-a543-09681539cf4c',
            'resource_id': 'fip-reservation-id-1',
            'resource_type': 'virtual:floatingip',
            'status': 'pending',
        }
        fip_reservation_get = self.patch(self.db_api, 'fip_reservation_get')
        fip_reservation_get.return_value = {
            'id': 'fip_resv_id1',
            'amount': 1,
            'reservation_id': '441c1476-9f8f-4700-9f30-cd9b6fef3509',
            'network_id': 'f548089e-fb3e-4013-a043-c5ed809c7a67',
        }
        self.assertRaises(mgr_exceptions.CantUpdateFloatingIPReservation,
                          fip_plugin.update_reservation,
                          '441c1476-9f8f-4700-9f30-cd9b6fef3509', values)

    def test_on_start(self):
        fip_reservation_get = self.patch(self.db_api, 'fip_reservation_get')
        fip_reservation_get.return_value = {
            'reservation_id': 'reservation-id1',
            'network_id': 'network-id1'
        }
        fip_allocation_get_all_by_values = self.patch(
            self.db_api, 'fip_allocation_get_all_by_values'
        )
        fip_allocation_get_all_by_values.return_value = [{
            'floatingip_id': 'fip-id',
        }]
        fip_get = self.patch(self.db_api, 'floatingip_get')
        fip_get.return_value = {
            'network_id': 'net-id',
            'subnet_id': 'subnet-id',
            'floating_ip_address': '172.2.24.100'
        }
        self.set_context(context.BlazarContext(project_id='fake-project-id'))
        m = mock.MagicMock()
        self.fip_pool.return_value = m

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        fip_plugin.on_start('resource-id1')

        self.fip_pool.assert_called_once_with('network-id1')
        m.create_reserved_floatingip.assert_called_once_with('subnet-id',
                                                             '172.2.24.100',
                                                             'fake-project-id',
                                                             'reservation-id1')

    def test_on_end(self):
        fip_reservation_get = self.patch(self.db_api, 'fip_reservation_get')
        fip_reservation_get.return_value = {
            'reservation_id': 'reservation-id1',
            'network_id': 'network-id1'
        }
        fip_allocation_get_all_by_values = self.patch(
            self.db_api, 'fip_allocation_get_all_by_values'
        )
        fip_allocation_get_all_by_values.return_value = [{
            'floatingip_id': 'fip-id',
        }]
        fip_get = self.patch(self.db_api, 'floatingip_get')
        fip_get.return_value = {
            'network_id': 'net-id',
            'subnet_id': 'subnet-id',
            'floating_ip_address': '172.2.24.100'
        }
        self.set_context(context.BlazarContext(project_id='fake-project-id'))
        m = mock.MagicMock()
        self.fip_pool.return_value = m

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        fip_plugin.on_end('resource-id1')

        self.fip_pool.assert_called_once_with('network-id1')
        m.delete_reserved_floatingip.assert_called_once_with('172.2.24.100')

    def test_matching_fips_not_allocated_fips(self):
        def fip_allocation_get_all_by_values(**kwargs):
            if kwargs['floatingip_id'] == 'fip1':
                return [{'id': 'allocation-id1'}]

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        fip_get = self.patch(self.db_api, 'reservable_fip_get_all_by_queries')
        fip_get.return_value = [
            {'id': 'fip1', 'floating_ip_address': '172.24.4.101'},
            {'id': 'fip2', 'floating_ip_address': '172.24.4.102'},
            {'id': 'fip3', 'floating_ip_address': '172.24.4.103'},
        ]
        fip_get = self.patch(self.db_api, 'fip_allocation_get_all_by_values')
        fip_get.side_effect = fip_allocation_get_all_by_values
        fip_get = self.patch(self.db_utils, 'get_free_periods')
        fip_get.return_value = [
            (datetime.datetime(2013, 12, 19, 20, 0),
             datetime.datetime(2013, 12, 19, 21, 0)),
        ]
        result = fip_plugin._matching_fips(
            'network-id', [], 2,
            datetime.datetime(2013, 12, 19, 20, 0),
            datetime.datetime(2013, 12, 19, 21, 0))
        self.assertEqual(['fip2', 'fip3'], result)

    def test_matching_fips_allocated_fips(self):
        def fip_allocation_get_all_by_values(**kwargs):
            return [{'id': kwargs['floatingip_id']}]

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        fip_get = self.patch(self.db_api, 'reservable_fip_get_all_by_queries')
        fip_get.return_value = [
            {'id': 'fip1', 'floating_ip_address': '172.24.4.101'},
            {'id': 'fip2', 'floating_ip_address': '172.24.4.102'},
            {'id': 'fip3', 'floating_ip_address': '172.24.4.103'},
        ]
        fip_get = self.patch(self.db_api, 'fip_allocation_get_all_by_values')
        fip_get.side_effect = fip_allocation_get_all_by_values
        fip_get = self.patch(self.db_utils, 'get_free_periods')
        fip_get.return_value = [
            (datetime.datetime(2013, 12, 19, 20, 0),
             datetime.datetime(2013, 12, 19, 21, 0)),
        ]
        result = fip_plugin._matching_fips(
            'network-id', [], 3,
            datetime.datetime(2013, 12, 19, 20, 0),
            datetime.datetime(2013, 12, 19, 21, 0))
        self.assertEqual(['fip1', 'fip2', 'fip3'], result)

    def test_matching_fips_allocated_fips_with_required(self):
        def fip_allocation_get_all_by_values(**kwargs):
            if kwargs['floatingip_id'] == 'fip1':
                return [{'id': 'allocation-id1'}]

        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        fip_get = self.patch(self.db_api, 'reservable_fip_get_all_by_queries')
        fip_get.return_value = [
            {'id': 'fip1', 'floating_ip_address': '172.24.4.101'},
            {'id': 'fip2', 'floating_ip_address': '172.24.4.102'},
            {'id': 'fip3', 'floating_ip_address': '172.24.4.103'},
            {'id': 'fip4', 'floating_ip_address': '172.24.4.104'},
        ]
        fip_get = self.patch(self.db_api, 'fip_allocation_get_all_by_values')
        fip_get.side_effect = fip_allocation_get_all_by_values
        fip_get = self.patch(self.db_utils, 'get_free_periods')
        fip_get.return_value = [
            (datetime.datetime(2013, 12, 19, 20, 0),
             datetime.datetime(2013, 12, 19, 21, 0)),
        ]
        result = fip_plugin._matching_fips(
            'network-id', ['172.24.4.102'], 4,
            datetime.datetime(2013, 12, 19, 20, 0),
            datetime.datetime(2013, 12, 19, 21, 0))
        # The order must be 1. required fips, 2. non-allocated fips,
        # then 3. allocated fips
        self.assertEqual(['fip2', 'fip3', 'fip4', 'fip1'], result)

    def test_matching_fips_allocated_fips_with_cleaning_time(self):
        def fip_allocation_get_all_by_values(**kwargs):
            return [{'id': kwargs['floatingip_id']}]

        self.cfg.CONF.set_override('cleaning_time', '5')
        fip_get = self.patch(
            self.db_api,
            'reservable_fip_get_all_by_queries')
        fip_get.return_value = [
            {'id': 'fip1', 'floating_ip_address': '172.24.4.101'},
            {'id': 'fip2', 'floating_ip_address': '172.24.4.102'},
            {'id': 'fip3', 'floating_ip_address': '172.24.4.103'},
        ]
        fip_get = self.patch(
            self.db_api,
            'fip_allocation_get_all_by_values')
        fip_get.side_effect = fip_allocation_get_all_by_values
        fip_get = self.patch(
            self.db_utils,
            'get_free_periods')
        fip_get.return_value = [
            (datetime.datetime(2013, 12, 19, 20, 0)
             - datetime.timedelta(minutes=5),
             datetime.datetime(2013, 12, 19, 21, 0)
             + datetime.timedelta(minutes=5))
        ]
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        result = fip_plugin._matching_fips(
            'network-id', [], 3,
            datetime.datetime(2013, 12, 19, 20, 0),
            datetime.datetime(2013, 12, 19, 21, 0))
        self.assertEqual(['fip1', 'fip2', 'fip3'], result)
        start_mergin = (datetime.datetime(2013, 12, 19, 20, 0)
                        - datetime.timedelta(minutes=5))
        end_mergin = (datetime.datetime(2013, 12, 19, 21, 0)
                      + datetime.timedelta(minutes=5))
        calls = [mock.call(fip, start_mergin, end_mergin,
                           end_mergin - start_mergin,
                           resource_type='floatingip')
                 for fip in ['fip1', 'fip2', 'fip3']]
        fip_get.assert_has_calls(calls)

    def test_matching_fips_not_matching(self):
        fip_plugin = floatingip_plugin.FloatingIpPlugin()
        fip_get = self.patch(
            self.db_api,
            'reservable_fip_get_all_by_queries')
        fip_get.return_value = []
        self.assertRaises(mgr_exceptions.NotEnoughFloatingIPAvailable,
                          fip_plugin._matching_fips,
                          'network-id', [], 2,
                          datetime.datetime(2013, 12, 19, 20, 0),
                          datetime.datetime(2013, 12, 19, 21, 0))
